import os

from str2bool import str2bool

GRPC_RETRY_ATTEMPTS = int(os.getenv("GRPC_RETRY_ATTEMPTS", "5"))
GRPC_RETRY_TIMEOUT = float(os.getenv("GRPC_RETRY_TIMEOUT", "60"))
GRPC_RETRY_DELAY = float(os.getenv("GRPC_RETRY_DELAY", "1"))
NUMEROUS_PROJECT = os.getenv("NUMEROUS_PROJECT")
NUMEROUS_SCENARIO = os.getenv("NUMEROUS_SCENARIO")
JOB_ID = os.getenv("JOB_ID")
NUMEROUS_EXECUTION_ID = os.getenv("NUMEROUS_EXECUTION_ID")
FORCE_INSECURE = str2bool(os.getenv("FORCE_INSECURE", "False"))
NUMEROUS_API_SERVER = os.getenv("NUMEROUS_API_SERVER")
NUMEROUS_API_PORT = (
    int(os.environ["NUMEROUS_API_PORT"]) if "NUMEROUS_API_PORT" in os.environ else None
)
NUMEROUS_CLEAR_DATA = str2bool(os.getenv("NUMEROUS_CLEAR_DATA", "False"))
NUMEROUS_API_REFRESH_TOKEN = os.getenv("NUMEROUS_API_REFRESH_TOKEN")

GRPC_MAX_MESSAGE_SIZE = int(os.environ.get("GRPC_MAX_MESSAGE_SIZE", 2**22))

DATA_STREAM_DEFAULT_TIMEOUT_SLEEP_FOR = float(
    os.environ.get("DATA_STREAM_DEFAULT_TIMEOUT_SLEEP_FOR", 0.5)
)



"""Authorization and authentication related functionality."""

import grpc

from numerous.sdk.connect.access_token import RefreshingAccessToken


class AccessTokenAuthMetadataPlugin(grpc.AuthMetadataPlugin):
    """Implements access-token based authorization for a gRPC channel."""

    def __init__(self, access_token: RefreshingAccessToken):
        self._access_token = access_token

    def __call__(
        self,
        context: grpc.AuthMetadataContext,  # noqa: F841
        callback: grpc.AuthMetadataPluginCallback,
    ):
        callback([("token", self._access_token.value), ("instance", "dummy")], None)

    def __eq__(self, other):
        if isinstance(other, AccessTokenAuthMetadataPlugin):
            return other._access_token == self._access_token
        else:
            return False



import threading
from enum import Enum
from time import sleep

from numerous.grpc import spm_pb2
from numerous.grpc.spm_pb2_grpc import SPMStub

from numerous.client_common.grpc_retry import grpc_retry_unary
from numerous.sdk.connect.config import Config
from numerous.sdk.connect.job_utils import JobIdentifier


class JobStatus(str, Enum):
    INITIALIZING = "initializing"
    RUNNING = "running"
    STOPPED = "stopped"
    FINISHED = "finished"
    ERROR = "error"
    FAILED = "failed"
    HIBERNATING = "hibernating"


class StatusHandler:
    """Handles job status updates, so they are not done too frequently."""

    def __init__(
        self,
        spm_client: SPMStub,
        identity: JobIdentifier,
        config: Config,
    ) -> None:
        self._spm_client = spm_client
        self._identity = identity
        self._running = True
        self._minimum_update_interval = config.min_status_update_interval_seconds
        self._max_status_message_length = config.max_status_message_length
        self._status: JobStatus = JobStatus.INITIALIZING
        self._progress: float = 0.0
        self._message: str = ""
        self._status_updater_continue = threading.Event()
        self._status_updater_thread = threading.Thread(
            name="StatusUpdater", target=self._status_updater, daemon=True
        )
        self._status_updater_thread.start()

    def close(self):
        self._running = False
        self._status_updater_continue.set()
        self._status_updater_thread.join()

    def _status_updater(self) -> None:
        self._set_scenario_progress()
        while self._running:
            sleep(self._minimum_update_interval)
            self._status_updater_continue.wait()
            self._status_updater_continue.clear()
            self._set_scenario_progress()

    @property
    def status(self) -> JobStatus:
        return self._status

    @status.setter
    def status(self, value: JobStatus) -> None:
        if self._status == value:
            return
        self._status = value
        self._status_updater_continue.set()

    @property
    def message(self) -> str:
        return self._message

    @message.setter
    def message(self, value: str) -> None:
        truncated = value[: self._max_status_message_length]
        if truncated == self._message:
            return
        self._message = truncated
        self._status_updater_continue.set()

    @property
    def progress(self) -> float:
        return self._progress

    @progress.setter
    def progress(self, value: float) -> None:
        clamped = _clamp(value, 0.0, 100.0)
        if clamped == self._progress:
            return
        self._progress = clamped
        self._status_updater_continue.set()

    @grpc_retry_unary()
    def _set_scenario_progress(self):
        self._spm_client.SetScenarioProgress(
            spm_pb2.ScenarioProgress(
                project=self._identity.project_id,
                scenario=self._identity.scenario_id,
                job_id=self._identity.job_id,
                status=self._status.value,
                progress=self._progress,
                message=self._message,
            )
        )


def _clamp(value: float, minimum: float, maximum: float):
    """Clamps the given value, between the given minimum and maximum values.

    :param minimum: The result will never be lower than the minimum.
    :param maximum: The result will never be higher than the maximum.
    """
    return min(max(minimum, value), maximum)



import enum
import json
import logging
import urllib.request
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import timedelta
from typing import IO, Any, Generator, Optional, Union

from numerous.grpc.spm_pb2 import FilePath, FilePaths
from numerous.grpc.spm_pb2_grpc import FileManagerStub

logger = logging.getLogger(__name__)


@dataclass
class Parameter:  # will be used as a base-class later for implementing specific functionality
    id: str
    uuid: str
    display_name: str
    type: str
    value: Any

    @staticmethod
    def from_document(
        data: dict[str, Any], file_manager_client: FileManagerStub
    ) -> "Parameter":
        return Parameter(
            id=data["id"],
            uuid=data["uuid"],
            display_name=data["displayName"],
            type=data["type"],
            value=parse_value(data, file_manager_client=file_manager_client),
        )


@dataclass
class FileReference:
    """Represents the reference to a selected file, uploaded to the platform. Can be
    used to access that files content."""

    id: str
    path: str
    name: str
    file_manager_client: FileManagerStub
    _downloaded_filepath: Optional[str] = field(default=None, init=False)

    @staticmethod
    def from_document(
        data: dict[str, Any], file_manager_client: FileManagerStub
    ) -> "FileReference":
        return FileReference(
            data["id"],
            data["path"],
            data["name"],
            file_manager_client=file_manager_client,
        )

    def _get_file_url(self) -> str:
        return (
            self.file_manager_client.GetDownloadUrls(
                request=FilePaths(
                    file_paths=[FilePath(path=self.path, content_type="text/plain")]
                )
            )
            .urls[0]
            .url
        )

    @property
    def file_path(self) -> str:
        """The local path to the referenced file. The file will be downloaded on access."""
        return (
            self._downloaded_filepath if self._downloaded_filepath else self.download()
        )

    def download(self) -> str:
        """Download the referenced file, and get the resulting path.

        If the file has already been downloaded, the previously downloaded file is
        overridden."""
        try:
            url = self._get_file_url()
            filepath, _ = urllib.request.urlretrieve(url)  # nosec
            self._downloaded_filepath = filepath
            return filepath
        except Exception as e:
            logger.exception(f"Error while downloading file: {e}")
            raise

    @contextmanager
    def open_text(self) -> Generator[IO[str], None, None]:
        """Open the referenced file for reading text content."""
        with open(self.file_path, "r") as f:
            yield f

    def read_json(self) -> Union[dict, list]:
        """Read the JSON content of the referenced file."""
        with self.open_text() as f:
            return json.load(f)

    def read_text(self) -> str:
        """Read the text content of the referenced file."""
        with self.open_text() as f:
            return f.read()

    @contextmanager
    def open_bytes(self) -> Generator[IO[bytes], None, None]:
        """Open the referenced file for reading binary content."""
        with open(self.file_path, "rb") as f:
            yield f

    def read_bytes(self) -> bytes:
        """Read the bytes of the referenced file."""
        with self.open_bytes() as f:
            return f.read()


class SimTimeUnit(enum.Enum):
    years = "years"
    months = "months"
    days = "days"
    hours = "hours"
    minutes = "minutes"
    seconds = "seconds"

    @classmethod
    def from_string(cls, value_str) -> "SimTimeUnit":
        for member in cls:
            if member.value == value_str:
                return member
        raise ValueError(f"No matching enum member for value: {value_str}")


@dataclass
class Currency:
    name: str
    code: str

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Currency":
        return Currency(data["name"], data["code"])


@dataclass
class Month:
    value: int

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Month":
        return Month(data["value"])


@dataclass
class MapLocation:
    id: str
    name: str
    center: tuple[float, float]

    @staticmethod
    def from_document(data: dict[str, Any]) -> "MapLocation":
        return MapLocation(data["id"], data["name"], data["center"])


@dataclass
class TimeValue:
    total_seconds: float
    unit: SimTimeUnit
    delta: timedelta

    @staticmethod
    def from_document(data: dict[str, Any]) -> "TimeValue":
        return TimeValue(
            float(data["value"]),
            SimTimeUnit.from_string(data["unit"]),
            timedelta(seconds=float(data["value"])),
        )


@dataclass
class Config:
    value: str

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Config":
        return Config(data["value"])


@dataclass
class Selection:
    id: str
    value: Union[str, float, int]

    @staticmethod
    def from_document(data: dict[str, Any]) -> "Selection":
        return Selection(data["tag"], data["value"])


def parse_value(data: dict[str, Any], file_manager_client: FileManagerStub) -> Any:
    value = data["value"]
    type_ = data["type"]
    if type_ == "string":
        return value
    elif type_ == "number":
        return float(value)
    elif type_ == "boolean":
        return value
    elif type_ == "file":
        return FileReference.from_document(value, file_manager_client)
    elif type_ == "json":
        return json.loads(value)
    elif type_ == "time_value":
        return TimeValue.from_document(value)
    elif type_ == "currency":
        return Currency.from_document(value)
    elif type_ == "month":
        return Month.from_document(value)
    elif type_ == "map":
        return MapLocation.from_document(value)
    elif type_ == "config":
        return Config.from_document(value)
    elif type_ == "selecter":
        return Selection.from_document(data)
    raise ValueError(f"Unkwnown parameter type: {type_}")


def get_parameters_dict(
    data: list[dict[str, Any]], file_manager_client: FileManagerStub
) -> dict[str, Parameter]:
    return {
        param.id: param
        for param in get_parameters_list(data, file_manager_client=file_manager_client)
    }


def get_parameters_list(
    data: list[dict[str, Any]], file_manager_client: FileManagerStub
) -> list[Parameter]:
    return [Parameter.from_document(param, file_manager_client) for param in data]



import io
import sys
from logging import Logger
from pathlib import Path
from typing import Optional

import requests

from ..auth import AuthenticationError, login
from ..build_manager_client import get_build_manager_client
from ..repository import NumerousRepository, ProjectScenario
from ..utils import (
    bold,
    cyan,
    extract_gzipped_tarball,
    green,
    hash_directory,
    print_load_warnings,
    red,
    yellow,
)


def command_checkout(
    log: Logger,
    path: Path,
    scenario: Optional[ProjectScenario],
    commit_id: Optional[str],
):
    path = Path.cwd() if path is None else path
    if not path.exists():
        print(red(f"Cannot checkout: {bold(path)} does not exist"))
        sys.exit(1)

    repo = NumerousRepository(path)
    try:
        _, load_warnings = repo.load()
    except Exception:
        print(red(f"Cannot checkout: {path} is not a repository."))
        sys.exit(1)

    print_load_warnings(load_warnings)

    if repo.remote is None:
        print(red("Cannot checkout: No remote is configured for the repository."))
        print(
            f"Use the command {bold('numerous config --remote <REMOTE_URL>')} to configure a remote for the "
            "repository."
        )
        sys.exit(1)

    try:
        id_token = login(repo)
    except AuthenticationError:
        print(red("Cannot checkout: Login failed."))
        sys.exit(1)

    with get_build_manager_client(
        repo.remote.api_url, id_token, repo.remote.organization
    ) as build_manager:
        if commit_id is None:
            if scenario is not None:
                scenario_id = scenario.id
            elif repo.scenario is not None:
                scenario_id = repo.scenario.id
            else:
                print(red("Cannot checkout: No scenario specified."))
            updates = build_manager.get_scenario_history(repo.remote.name, scenario_id)
            update = updates[0] if updates else None
        else:
            update = build_manager.get_history_update(repo.remote.name, commit_id)

        if update is None:
            if scenario:
                repo.scenario = scenario
                repo.save()
                print(green(f"Checked out fresh scenario {bold(scenario)}"))
                return
            else:
                print(red("Cannot checkout: no commit found"))
                sys.exit(1)

        snapshot = hash_directory(path)
        choice = input(
            yellow("There are changes since the last commit. Continue? (y/n) ")
        )  # nosec
        if snapshot != repo.snapshot and choice.strip().lower() != "y":
            print(red("Aborted checkout"))
            sys.exit(1)

        print(
            cyan(f"Downloading snapshot {bold(update.snapshot_id)} into {bold(path)}")
        )
        url = build_manager.get_download_url(
            repo.remote.name, update.project_id, update.scenario_id, update.snapshot_id
        )
        response = requests.get(url, allow_redirects=True)
        extract_gzipped_tarball(io.BytesIO(response.content), path)

        repo.snapshot = update.snapshot_id
        repo.commit = update.commit_id
        repo.scenario = ProjectScenario(
            scenario_id=update.scenario_id, project_id=update.project_id
        )
        repo.save()

        print(
            green(
                f"Checkout complete => Scenario {bold(repo.scenario)}, commit {bold(repo.commit)}"
            )
        )



import logging
import os
from datetime import datetime
from enum import Enum
from time import sleep, time
from typing import Any, Optional, Union

from numerous.client import ScenarioStatus
from numerous.client.data_source import (
    DataSourceCompleted,
    DataSourceEmptyDataset,
    DataSourceHibernating,
    DataSourceStreamClosed,
    DataSourceStreamStarting,
)

from .errors import SimulationError, tb_str
from .system import NumerousSystem

LOG_LEVEL = os.getenv("LOG_LEVEL", logging.DEBUG)


class ExitCode(Enum):
    def __new__(cls, *args, **kwds):  # noqa: F841
        value = args[0]
        obj = object.__new__(cls)
        obj._value_ = value
        return obj

    def __init__(self, _, description, scenariostatus):
        self.description = description
        self.scenariostatus = scenariostatus

    COMPLETED = 10, "Scenario completed", ScenarioStatus.FINISHED
    SOURCE_COMPLETED = 11, "Scenario source completed", ScenarioStatus.FINISHED
    TERMINATED = 20, "Scenario terminated", ScenarioStatus.FINISHED
    HIBERNATING = 30, "Scenario hibernating", ScenarioStatus.HIBERNATING
    SOURCE_HIBERNATING = 31, "Scenario source hibernating", ScenarioStatus.HIBERNATING
    SOURCE_CLOSED = -20, "Scenario source closed", ScenarioStatus.FINISHED
    TIMEOUT = -30, "Data could not be retrieved before timeout", ScenarioStatus.FAILED
    ERROR = -100, "An error occured. See logs.", ScenarioStatus.FAILED
    SOURCE_EMPTY = -110, "Scenario source empty", ScenarioStatus.FINISHED
    SYSTEM_NOT_INITIALIZED = (
        -120,
        "System could not be initialized",
        ScenarioStatus.FAILED,
    )


class NumerousBaseJob:
    def __init__(self):
        self.system = None
        self.input = None
        self.app = None
        self.logger = logging.getLogger("numerous-base-job")
        self.init = True
        self.last_print = {}
        self._t0 = None
        self._exit_code = None
        self._t = None
        self._details = ""
        # Wrapper to allow specialized instanciation of components
        self.wrapper = None

    def read_data(self, t):
        if self.app.client.terminate_event.is_set():
            raise TerminateJob
        if self.app.client.hibernate_event.is_set():
            raise HibernateJob

        while True:
            try:
                data = self.input.get_at_time(t)
            except DataSourceStreamStarting:
                self.app.client.set_scenario_progress(
                    "waiting for input stream to start", ScenarioStatus.INITIALZING, 0.0
                )
                sleep(0.1)
                continue
            except Exception:
                raise

            self.system.update_inputs(data)

            return data

    def _post_init(self):  # This is called before _run_job from the app.loop method
        timeout = 120
        t, states = self._load_states()
        dt = self.app.client.params.get("dt_simulation", 60)

        # Allow dt parameter to be a time value from front-end
        if isinstance(dt, dict):
            dt = dt["value"]

        t_stop = self.app.end_time if not self.app.subscribe else 0
        self._t0 = t

        self.input = self.app.client.get_inputs(
            self.app.scenario, t0=t, te=t_stop, dt=dt, tag_prefix="", timeout=timeout
        )

        self.system = NumerousSystem(
            self.app.client,
            self.app.scenario,
            self.app.files,
            self.app.start_time,
            t_stop,
            self.app.model_folder,
            self,
            states,
            t,
            dt,
            self.input,
            wrapper=self.wrapper,
        )

    def serialize_states(self, t: float) -> Optional[Any]:
        """
        A method called when saving states.
        Returns: states, which must be JSON serializable, to be saved.
        """
        return {}

    def _save_states(self, t, cause, details):
        if self.init:
            self.logger.info("model not initialized, could not save states")
            return

        model_states = self.serialize_states(t)
        states = {"t": t, "states": model_states, "cause": cause, "details": details}

        self.app.client.state.set("states", states)

        self.logger.debug(f"states saved: reason {cause} ({details})")

    def _load_states(self):
        t = self.app.start_time
        states = self.app.client.state.get("states", {})

        t = states.get("t", t)

        if states:
            self.logger.debug("states loaded")
        return t, states

    def _print_update(self, t, ix, update_interval=5):
        if self.last_print.get(ix) is None:
            self.last_print.update({ix: t})
            return True

        if t - self.last_print[ix] >= update_interval:
            self.last_print.update({ix: t})
            return True
        else:
            return False

    def _start(self, app) -> ExitCode:
        self.app = app
        exit_code = ExitCode.COMPLETED
        try:
            self._post_init()
            exit_code = self.run_job()
            if not exit_code:
                exit_code = ExitCode.COMPLETED
        except Exception as e:
            self.logger.error(f"An error occurred {tb_str(e)}")
            exit_code = ExitCode.ERROR
        finally:
            self.logger.info(f"job ended with status {exit_code}")
            self._exit_code = exit_code
            self._save_states(self._t, exit_code.description, self._details)
            self.post_run(exit_code)
            return exit_code

    def run_job(self) -> ExitCode:
        # Overwrite this method to create the simplest job
        raise NotImplementedError

    def post_run(self, exit_code: ExitCode):
        pass


class TerminateJob(Exception):
    pass


class HibernateJob(Exception):
    pass


class NumerousSimulationJob(NumerousBaseJob):
    def __init__(self) -> None:
        super(NumerousSimulationJob, self).__init__()
        self.system: Optional[NumerousSystem] = None
        self.align_outputs_to_next_timestep = True
        self.logger = logging.getLogger("numerous-simulation-job")
        self.logger.setLevel(level=LOG_LEVEL)
        self._completed = 0
        self._t_last = None

    def write_output(self, t, output):
        if not output:
            raise SimulationError("Expected an output, but got None")

        if self.init:
            self.app.client.set_timeseries_meta_data(
                [{"name": tag} for tag in output.keys()], offset=self.app.start_time
            )
            self.init = False

        if not self.align_outputs_to_next_timestep:
            output.update({"_index": self._t - self.app.start_time})
        else:
            output.update({"_index": t - self.app.start_time})

        self.app.output.write_row(t, output)

    def read_input(self, t: float):
        while True:
            data = self.read_data(t)
            if not data:
                self.app.client.set_scenario_progress(
                    f"waiting for data. Last update: " f"{datetime.fromtimestamp(t)}",
                    ScenarioStatus.RUNNING if not self.init else ScenarioStatus.WAITING,
                    self._completed,
                )
                if self._print_update(datetime.now().timestamp(), 0, 10):
                    self.logger.info(f"no data. Simulation time: {t}. ")
                sleep(1)
                continue
            break

        if t == 0 and data["_index"] > 0:
            self.logger.warning(f'setting start time to {data["_index"]}')
            self._t = data["_index"]

        return None

    def initialize_simulation_system(self) -> Optional[dict]:
        """
        A method that is called once model is initialized (after the first data is read).
        Can return the initial output to be saved.
        """
        return None

    def step(self, t: float, dt: float) -> tuple[float, dict]:
        """
        A method that is called after each data read. Could be a step solver, or some other data manipulating function.
        Returns: tuple of next timestamp and outputs as a dict with tags to be saved

        """
        return t + dt, {"no_job_defined": None}

    def _handle_data_exception(self, e: Exception):
        if e.__class__ == DataSourceHibernating:
            if self.app.allow_hibernation:
                exit_code = ExitCode.SOURCE_HIBERNATING
                self.app.client.hibernate(message="hibernating")
            else:
                exit_code = ExitCode.COMPLETED
            return exit_code
        elif e.__class__ == DataSourceCompleted:
            return ExitCode.SOURCE_COMPLETED
        elif e.__class__ == DataSourceEmptyDataset:
            return ExitCode.SOURCE_EMPTY
        elif e.__class__ == DataSourceStreamClosed:
            return ExitCode.SOURCE_CLOSED
        elif e.__class__ == TimeoutError:
            return ExitCode.TIMEOUT
        elif e.__class__ == TerminateJob:
            return ExitCode.COMPLETED
        elif e.__class__ == HibernateJob:
            return ExitCode.HIBERNATING

    def run_job(self) -> ExitCode:
        if self.system is None:
            self.app.client.set_scenario_progress(
                "Could not load model",
                ScenarioStatus.FAILED,
                0.0,
                force=True,
            )
            return ExitCode.ERROR
        elif self.system.end_time is None:
            self.app.client.set_scenario_progress(
                "No end time set",
                ScenarioStatus.FAILED,
                0.0,
                force=True,
            )
            return ExitCode.ERROR

        # This is the classic pipelines approach - well suited for digital twins
        # Run simulation in a loop
        self.logger.debug("entering loop")
        details: Union[dict[str, Any], str] = ""
        last_data = {}

        t = None
        exit_code = ExitCode.COMPLETED
        try:
            self.app.client.set_scenario_progress(
                "bootup", ScenarioStatus.INITIALZING, 0.0, force=True
            )
            self.app.terminate.wait(timeout=5)
            self.app.client.set_scenario_progress(
                "waiting for initial data", ScenarioStatus.INITIALZING, 0.0, force=True
            )
            t_stop = self.system.end_time
            self._t = self._t0

            # Increasingly longer spacing between log updates to 60 s
            last_logged = -1000.0
            dt_log_max = 60.0
            dt_log = 1.0

            while True:
                if t_stop > 0:
                    self._completed = (
                        1 - (t_stop - self._t) / (t_stop - self.app.start_time)
                    ) * 100
                    self.app.client.set_scenario_progress(
                        "running", ScenarioStatus.RUNNING, self._completed
                    )
                else:
                    self.app.client.set_scenario_progress(
                        f"Simulation time: {datetime.fromtimestamp(self._t)}",
                        ScenarioStatus.RUNNING,
                        0,
                    )

                if self.app.backup.is_set():
                    self._save_states(
                        t, "backup", f"scheduled checkpoint @ {datetime.now()}"
                    )
                    self.app.backup.clear()

                self.read_input(self._t)

                if self.init:
                    self.app.client.set_scenario_progress(
                        "building model",
                        ScenarioStatus.MODEL_INITIALIZING,
                        0,
                        force=True,
                    )
                    initial_output = self.initialize_simulation_system()
                    if initial_output:
                        if "_index" not in initial_output:
                            initial_output.update({"_index": 0})
                        self.write_output(self._t, initial_output)

                try:
                    tnew, output = self.step(self._t, self.system.dt)
                    self.write_output(tnew, output)
                except Exception as e:
                    details = tb_str(e)
                    raise SimulationError(details)

                # Advance time
                self._t = tnew

                if time() - last_logged > dt_log:
                    last_logged = time()
                    dt_log = min(dt_log_max, dt_log * 2)

                    self.logger.info(
                        f"Calculation step. Time is now {self._t}. completed: {self._completed}"
                    )

                last_data = {
                    component.name: component.inputs
                    for component in self.system.components.values()
                }

                if (self._t >= t_stop) and (t_stop > 0):
                    self.logger.warning("maximum time reached")
                    exit_code = ExitCode.COMPLETED
                    break

        except (
            TimeoutError,
            TerminateJob,
            HibernateJob,
            DataSourceHibernating,
            DataSourceCompleted,
            DataSourceEmptyDataset,
            DataSourceStreamClosed,
        ) as e:
            exit_code = self._handle_data_exception(e)
            details = tb_str(e)

        except Exception as e:
            self.logger.error(f"numerous_app crashed: {tb_str(e)}")
            self.logger.debug(f"previous data: {last_data}")
            data = {
                component.name: component.inputs
                for component in self.system.components.values()
            }
            self.logger.debug(f"data at crash: {data}")
            details = {
                "error message": tb_str(e),
                "previous_data": last_data,
                "data at crash": data,
            }
            exit_code = ExitCode.ERROR
        finally:
            self.logger.info("job terminated")
            self._details = details
            return exit_code



"""Contains the implementation of the JobState class, for managing job execution state."""

import json
from typing import Any, Optional, Union

from numerous.grpc.spm_pb2 import ScenarioCustomMetaData
from numerous.grpc.spm_pb2_grpc import SPMStub

from numerous.client_common.grpc_retry import grpc_retry_unary
from numerous.sdk.connect.job_utils import JobIdentifier


class JobState:
    """JobState enables jobs to store state remotely, and cache it locally. This is useful for keeping track of values
    across hibernations, or if an error has restarted the job.
    """

    def __init__(
        self,
        spm_client: SPMStub,
        identity: JobIdentifier,
        execution_id: str,
    ) -> None:
        """Initializes the JobState.

        :param spm_client:
        :param identity: The identitiers for the job and related objects.
        """
        self._spm_client = spm_client
        self._identity = identity
        self._execution_id = execution_id
        self._state = self._load()

    def get(self, key: str, default: Optional[Any] = None) -> Union[Any, Any]:
        """Get the state value keyed by :paramref:`key` from the :class:`JobState`.

        :param key: The key to look up in the state.
        :param default: A default value to return if the key does not exist.
        :return: The state value at the :paramref:`key` if it exists, otherwise the
            :paramref:`default`, or `None` if no default is specified.
        """
        return self._state.get(key, default)

    def set(self, key: str, value: Any, commit: bool = True) -> None:
        """Set the state value keyed by :paramref:`key` to the value :paramref:`value`
        in the :class:`JobState`, and optionally commit the state to the server.

        If :paramref:`commit` is `True` the new state resulting from setting the value,
        will be comitted to the server.

        :param key: The key to set in the state.
        :param value: The value to set.
        :param commit: Whether or not to commit the resulting state to the server.
        """
        self._state[key] = value

        if commit:
            self.commit()

    def unset(self, key: str, commit: bool = True) -> None:
        """Unset the state value keyed by :paramref:`key` in the :class:`JobState`, and
        optionally commit the state to the server.

        If :paramref:`commit` is `True` the new state resulting from unsetting the
        value, will be comitted to the server.

        :param key: The key to unset in the state.
        :param commit: Whether or not to commit the resulting state to the server.
        """
        if key in self._state:
            del self._state[key]

        if commit:
            self.commit()

    @grpc_retry_unary()
    def _load(self) -> dict[str, Any]:
        try:
            state = json.loads(
                self._spm_client.GetScenarioCustomMetaData(
                    request=ScenarioCustomMetaData(
                        project=self._identity.project_id,
                        scenario=self._identity.scenario_id,
                        execution=self._execution_id,
                        key=f"{self._identity.job_id}.state",
                    )
                ).meta
            )
        except json.decoder.JSONDecodeError:
            state = {}

        if isinstance(state, dict):
            return state
        else:
            return {}

    @grpc_retry_unary()
    def commit(self) -> None:
        """Commit the state to the server. This is useful for batching state updates."""
        self._spm_client.SetScenarioCustomMetaData(
            request=ScenarioCustomMetaData(
                project=self._identity.project_id,
                scenario=self._identity.scenario_id,
                execution=self._execution_id,
                key=f"{self._identity.job_id}.state",
                meta=json.dumps(self._state),
            )
        )



"""numerous.sdk - develop numerous applications!

The recommended way to develop applications for running on the numerous platform.
"""
from numerous.sdk.connect.job_client import JobClient, JobStatus
from numerous.sdk.connect.job_state import JobState
from numerous.sdk.models.component import Component
from numerous.sdk.models.input import InputVariable
from numerous.sdk.models.job import Job
from numerous.sdk.models.job_time import JobTime, RunMode
from numerous.sdk.models.parameter import Parameter
from numerous.sdk.models.scenario import Scenario

__all__ = [
    "JobClient",
    "JobStatus",
    "JobState",
    "Component",
    "InputVariable",
    "Job",
    "Parameter",
    "Scenario",
    "RunMode",
    "JobTime",
]



import io
import mimetypes
import os
import tempfile
from typing import IO, Optional

import requests
from numerous.grpc.spm_pb2 import ScenarioFilePath
from numerous.grpc.spm_pb2_grpc import SPMStub


class FileManager:
    def __init__(self, spm_client: SPMStub, project_id: str, scenario_id: str):
        self._spm_client = spm_client
        self._project_id = project_id
        self._scenario_id = scenario_id

    def upload_bytes(
        self, file_name: str, content: bytes, content_type: Optional[str] = None
    ) -> None:
        """Uploads the given content with the given file name, as the content type if specified.

        :param file_name:
        :param content: The bytes file content.
        :param content_type: The file content media type.
        """
        buffer = io.BytesIO(content)
        self._upload(file_name, file_name, buffer, content_type)

    def upload_stream(
        self, file_name: str, stream: IO[bytes], content_type: Optional[str] = None
    ) -> None:
        """Uploads the content read from the given content stream.
        Uses the given file name, and content type, if specified.

        :param file_name:
        :param stream: The file content bytes stream.
        :param content_type: The file content media type.
        """
        self._upload(file_name, file_name, stream, content_type)

    def upload_text(self, file_name: str, content: str) -> None:
        """Uploads the content with the given file name, as plain text.

        :param file_name:
        :param content: The text file content.
        """
        buffer = io.BytesIO(content.encode("utf-8"))
        self._upload(file_name, file_name, buffer, "text/plain")

    def upload_path(self, path: str) -> None:
        """Uploads the file at the given path, inferring the content type.

        :param path:
        """
        with open(path, "rb") as f:
            self._upload(path, os.path.split(path)[-1], f)

    def _upload(
        self,
        path: str,
        file_id: str,
        data: IO[bytes],
        content_type: Optional[str] = None,
    ):
        if content_type is None:
            content_type = mimetypes.guess_type(path)[0] or "application/octet-stream"
        scenario_file_path = ScenarioFilePath(
            project=self._project_id,
            scenario=self._scenario_id,
            path=path,
            file_id=file_id,
            content_type=content_type,
        )
        upload_url = self._spm_client.GenerateScenarioUploadSignedURL(
            scenario_file_path
        )
        requests.post(
            upload_url.url,
            headers={"Content-Type": "multipart/related"},
            params={"name": path, "mimeType": content_type},
            data=data,
        )


class FileUploadContext:
    _chunk_size = 4096

    def __init__(self, file_manager: FileManager, file_name: str):
        """
        :param file_manager: A file manager which is used to upload the content.
        :param file_name: The output file name.
        """
        self._file_manager = file_manager
        self._file_name = file_name
        self._buffer: Optional[io.BytesIO] = None

    def __enter__(self):
        """Creates a buffer for the content to be uploaded."""
        self._buffer = tempfile.TemporaryFile(prefix="numerous_sdk")
        return self

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        """Executes actual content upload from a buffer."""
        if not self._buffer:
            raise RuntimeError("Buffer is not initialized")

        try:
            self._buffer.seek(0)
            self._file_manager.upload_stream(self._file_name, self._buffer)
        finally:
            self._buffer.close()

    def write(self, data: str):
        """Adds the content of a given string.

        :param data: The content.
        """
        if not self._buffer:
            raise RuntimeError("Buffer is not initialized")

        self._buffer.write(data.encode())

    def write_from_file(self, file_name: str):
        """Adds the content from a given file.

        :param file_name: the name of a file with the content.
        """
        if not self._buffer:
            raise RuntimeError("Buffer is not initialized")

        with open(file_name, "rb") as f:
            while buffer := f.read(self._chunk_size):
                self._buffer.write(buffer)



"""Components are the basic building blocks of systems.

This module contains the definition of Components, and the implementation of
functionality to create them.
"""
from collections import defaultdict
from dataclasses import dataclass
from typing import Any

from numerous.grpc.spm_pb2_grpc import FileManagerStub

from numerous.sdk.models.input import (
    InputSource,
    InputVariable,
    InputVariableScenarioSource,
    InputVariableStaticSource,
)
from numerous.sdk.models.parameter import Parameter, get_parameters_dict


@dataclass
class SubComponentNotFound(Exception):
    """Raised when creating a :class:`Component` is configured with a subcomponent that
    does not exist.
    """

    component_uuid: str


@dataclass
class Component:
    """A component of a :class:`~numerous.sdk.models.scenario.Scenario`, containing
    parameters, input variables, and any subcomponents."""

    uuid: str
    id: str
    type: str
    name: str
    item_class: list[str]
    is_enabled: bool
    is_main: bool
    input_variables: dict[str, InputVariable]
    components: dict[str, list["Component"]]
    parameters: dict[str, Parameter]
    """The subcomponents of this component."""

    @staticmethod
    def from_document(
        path: list[str],
        component: dict[str, Any],
        all_components: list[dict[str, Any]],
        input_sources_mapping: dict[str, InputSource],
        file_manager_client: FileManagerStub,
    ) -> "Component":
        component_path = path + [component["name"]]
        return Component(
            uuid=component["uuid"],
            id=component["id"],
            type=component["type"],
            name=component["name"],
            is_enabled=not bool(component["disabled"]),
            is_main=bool(component["isMainComponent"]),
            components=_extract_subcomponents(
                component_path,
                component,
                all_components,
                input_sources_mapping,
                file_manager_client,
            ),
            item_class=component["item_class"].split("."),
            input_variables=_extract_input_variables(
                component_path, component, input_sources_mapping
            ),
            parameters=get_parameters_dict(
                component.get("parameters", []), file_manager_client
            ),
        )


def _extract_subcomponents(
    path: list[str],
    component: dict[str, Any],
    all_components: list[dict[str, Any]],
    input_sources_mapping: dict[str, InputSource],
    file_manager_client: FileManagerStub,
) -> dict[str, list[Component]]:
    subcomponents = defaultdict(list)
    for subcomponent_ref in component.get("subcomponents", []):
        subcomponents[subcomponent_ref["name"]].append(
            _extract_subcomponent(
                path,
                subcomponent_ref["uuid"],
                all_components,
                input_sources_mapping,
                file_manager_client,
            )
        )
    return subcomponents


def _extract_subcomponent(
    path: list[str],
    subcomponent_uuid: str,
    all_components: list[dict[str, Any]],
    input_sources_mapping: dict[str, InputSource],
    file_manager_client: FileManagerStub,
) -> Component:
    subcomponent = next(
        (
            subcomponent
            for subcomponent in all_components
            if subcomponent.get("uuid") == subcomponent_uuid
        ),
        None,
    )

    if subcomponent is None:
        raise SubComponentNotFound(component_uuid=subcomponent_uuid)

    return Component.from_document(
        path,
        subcomponent,
        all_components,
        input_sources_mapping=input_sources_mapping,
        file_manager_client=file_manager_client,
    )


def extract_components(
    data: dict[str, Any],
    input_sources_mapping: dict[str, InputSource],
    file_manager_client,
) -> dict[str, Component]:
    components = data.get("simComponents", [])
    subcomponent_uuids = _get_all_subcomponent_uuids(components)
    return {
        component["name"]: Component.from_document(
            [], component, components, input_sources_mapping, file_manager_client
        )
        for component in components
        if component["uuid"] not in subcomponent_uuids
    }


def _get_all_subcomponent_uuids(components: list[dict[str, Any]]) -> set[str]:
    all_subcomponent_uuids = set()

    for component in components:
        for subcomponent in component.get("subcomponents", []):
            all_subcomponent_uuids.add(subcomponent["uuid"])

    return all_subcomponent_uuids


def _extract_input_variable(
    path: list[str], data: dict[str, Any], input_sources_mapping: dict[str, InputSource]
) -> InputVariable:
    input_variable_path = path + [data["id"]]

    input_variable = InputVariable(
        id=data["id"],
        uuid=data["uuid"],
        path=input_variable_path,
        display_name=data["display"],
        source=_extract_input_variable_source(data, input_sources_mapping),
        scale=data["scaling"],
        offset=data["offset"],
    )
    return input_variable


def _extract_input_variable_source(
    input_variable: dict[str, Any], input_sources_mapping: dict[str, InputSource]
):
    data_source_type = input_variable["dataSourceType"]
    if data_source_type == "static":
        return InputVariableStaticSource(value=input_variable["value"])
    elif data_source_type in ["scenario", "control_machines"]:
        data_source_id = input_variable["dataSourceID"]
        tag_source = input_variable["tagSource"]
        input_source = input_sources_mapping[data_source_id]
        if (
            tag_source["projectID"] != input_source.project_id
            or tag_source["scenarioID"] != input_source.scenario_id
        ):
            raise ValueError("Tag source does not match selected input source")
        return InputVariableScenarioSource(input_source, tag=tag_source["tag"])
    else:
        raise ValueError(f"Invalid data source type: {data_source_type}")


def _extract_input_variables(
    path: list[str],
    component: dict[str, Any],
    input_sources_mapping: dict[str, InputSource],
) -> dict[str, InputVariable]:
    variables = component.get("inputVariables", [])
    variables_list: list[InputVariable] = [
        _extract_input_variable(path, variable, input_sources_mapping)
        for variable in variables
    ]
    return {variable.id: variable for variable in variables_list}



"""Functionality related to writing data series and the :class:`Writer`."""


import sys
from collections import defaultdict
from datetime import datetime
from typing import Any, Callable, Generator, Iterable, Optional, Sequence, Union

from numerous.grpc import spm_pb2, spm_pb2_grpc

from numerous.sdk.connect.job_utils import JobIdentifier
from numerous.sdk.connect.request_response_stream import RequestResponseStream

FLOAT_SIZE = sys.getsizeof(float())


class TagsNotAllowed(Exception):
    """An error raised when written tags are invalid according to the data series."""

    def __init__(self, tags: Iterable[str]):
        """Initialize the error.

        :param tags: The unallowed tags which triggered the error.
        """
        self.tags = tags

    def __eq__(self, __o: object) -> bool:  # pragma: no cover
        return isinstance(__o, TagsNotAllowed) and __o.tags == self.tags


def _default_tag(name: str):
    return spm_pb2.Tag(
        name=name,
        type="double",
        scaling=1,
        offset=0,
    )


class Writer:
    """The :class:`Writer` enables writing rows and series to the server.
    It manages a buffer, which is automatically flushed when it is full.
    """

    def __init__(
        self,
        spm_stub: spm_pb2_grpc.SPMStub,
        job_identity: JobIdentifier,
        start_time: datetime,
        execution_id: str,
        max_size_bytes: int,
        flush_margin_bytes: int,
    ):
        """Initialize the writer.

        :param spm_stub: The client.
        :param job_identity: The identity of the job, the :class:`Writer` is writing for.
        :param execution_id: The ID of the execution, the :class:`Writer` is writing for.
        :param max_size_bytes: The maximum buffer size of the :class:`Writer` in bytes.
        :param flush_margin_bytes: The margin that is subtracted the
        :paramref:`max_size_bytes` to determine if flushing should be performed.
        """
        self._spm_stub = spm_stub
        self._job_identity = job_identity
        self._start_time = start_time
        self._execution_id = execution_id
        self._buffer_index: list[float] = []
        self._buffer: dict[str, list[float]] = defaultdict(list)
        self._max_size_bytes = max_size_bytes
        self._flush_margin_bytes = flush_margin_bytes
        self._size_bytes: int = 0
        write_data_stream: Callable[
            [Iterable[spm_pb2.WriteDataStreamRequest]],
            Generator[spm_pb2.WriteDataStreamResponse, None, None],
        ] = spm_stub.WriteDataStream  # type: ignore[assignment]
        self._stream = RequestResponseStream(write_data_stream)
        self._allowed_tags: Optional[set[str]] = None
        self._closed: bool = False

    def _get_allowed_tags(self) -> Optional[set[str]]:
        if self._allowed_tags is not None:
            return self._allowed_tags

        metadata = self._spm_stub.GetScenarioMetaData(
            spm_pb2.Scenario(
                project=self._job_identity.project_id,
                scenario=self._job_identity.scenario_id,
                execution=self._execution_id,
            )
        )
        if metadata.tags:
            self._allowed_tags = {tag.name for tag in metadata.tags}
            return self._allowed_tags
        else:
            return None

    def _get_unallowed_tags(
        self, data: dict[str, Any], tags: Iterable[str]
    ) -> set[str]:
        return set(data.keys()).difference(tags)

    def _header_size(self, tags: Iterable[str]) -> int:
        if self._size_bytes == 0:
            return sum(sys.getsizeof(tag) for tag in tags)
        else:
            return 0

    def _list_size(self, data: Sequence[Union[float, datetime]]) -> int:
        return FLOAT_SIZE * len(data)

    def _row_and_index_size(self, data: dict[str, float]) -> int:
        return FLOAT_SIZE + FLOAT_SIZE * len(data)

    def _series_size(
        self, index: Sequence[Union[datetime, float]], data: dict[str, list[float]]
    ) -> int:
        return self._list_size(index) + sum(
            (self._list_size(values) for values in data.values())
        )

    def _must_flush_before_adding(self, added_size_bytes: int) -> bool:
        return (
            added_size_bytes + self._size_bytes + self._flush_margin_bytes
            >= self._max_size_bytes
        )

    def _set_data_series_config(self, data: dict[str, Any]) -> None:
        self._allowed_tags = set(data.keys())
        self._spm_stub.SetScenarioMetaData(
            spm_pb2.ScenarioMetaData(
                offset=self._start_time.timestamp(),
                tags=[_default_tag(tag) for tag in data.keys()],
                project=self._job_identity.project_id,
                scenario=self._job_identity.scenario_id,
                execution=self._execution_id,
            )
        )

    def _check_data_series_config(self, data: dict[str, Any]):
        if (tags := self._get_allowed_tags()) is None:
            self._set_data_series_config(data)
        elif unallowed_tags := self._get_unallowed_tags(data, tags):
            raise TagsNotAllowed(unallowed_tags)

    def row(
        self, time: Union[float, datetime], data: dict[str, float], flush: bool = False
    ) -> None:
        """Write a row of data into the data series.

        Each row written to the :class:`Writer` must have same keys as previously
        written series or rows.

        :param time: The time, at which the data is indexed. If a `float` value, it is
            interpreted as an offset in seconds since the start time of the job,
            otherwise if it is a `datetime`, it is interpreted as the time within the
            job runtime.
        :param data: The row of data to write into the data stream.
        :param flush: If true, will flush the buffer of the writer upon writing.
        :raises TagsNotAllowed: If tags that have not been previously written, are
            written.

        Writing rows can be used alongside :meth:`numerous.sdk.JobClient.read_inputs`,
        which will also iterate over the jobs runtime:

            >>> for t, data in job_client.read_inputs(step=3600.0):
            ...     job_client.writer.row(t, {"output": data["input1"] * 2})

        The type of the time values does not *have* to be the same type for each call:

            >>> job_client.job_time.start
            datetime(2023, 6, 1, 12, 0 ,0)
            >>> job_client.writer.row(datetime(2023, 6, 1, 12, 0, 0), {"value": 1.0})
            >>> job_client.writer.row(3600.0, {"value": 2.0})
            >>> job_client.writer.row(datetime(2023, 6, 1, 14, 0, 0), {"value": 4.0})
        """
        self._check_data_series_config(data)

        size_bytes = self._header_size(data) + self._row_and_index_size(data)
        if self._must_flush_before_adding(size_bytes):
            self.flush()

        index = self._time_to_index(time)
        for tag, value in data.items():
            self._buffer[tag].append(value)
        self._buffer_index.append(index)
        self._size_bytes += size_bytes

        if flush:
            self.flush()

    def series(
        self,
        times: Sequence[Union[float, datetime]],
        data: dict[str, list[float]],
        flush: bool = False,
    ) -> None:
        """Write a series of data to the data stream.

        Each series written to the :class:`Writer` must have same keys as previously
        written series or rows.

        :param times: The time values of the series, like for :meth:`row`, they
            can be either offsets from the start time, or a point in time.
        :param data: The data series to write. The keys are validated against
            previously written keys.
        :param flush: If true, will flush the buffer of the writer upon writing.
        :raises TagsNotAllowed: If tags that have not been previously written, are
            written.
        """
        self._check_data_series_config(data)

        size_bytes = self._header_size(data) + self._series_size(times, data)
        if self._must_flush_before_adding(size_bytes):
            self.flush()

        for key, value in data.items():
            self._buffer[key].extend(value)
        for time in times:
            self._buffer_index.append(self._time_to_index(time))
        self._size_bytes += size_bytes

        if flush:
            self.flush()

    def _time_to_index(self, time: Union[float, datetime]) -> float:
        return (
            (time - self._start_time).total_seconds()
            if isinstance(time, datetime)
            else time
        )

    def flush(self) -> None:
        """Flushes the buffer of the writer."""
        self._stream.send(
            spm_pb2.WriteDataStreamRequest(
                scenario=self._job_identity.scenario_id,
                execution=self._execution_id,
                overwrite=False,
                index=self._buffer_index,
                data={
                    tag: spm_pb2.StreamData(values=values)
                    for tag, values in self._buffer.items()
                },
                update_stats=True,
            )
        )
        self._buffer_index.clear()
        self._buffer.clear()
        self._size_bytes = 0

    def close(self, hibernating: bool):
        """Close the :class:`Writer`, flushing the buffer."""
        if not self._closed:
            self.flush()
            self._stream.close()
            self._spm_stub.CloseData(
                spm_pb2.DataCompleted(
                    scenario=self._job_identity.scenario_id,
                    execution=self._execution_id,
                    finalized=not hibernating,
                )
            )
            self._closed = True



import importlib.metadata

from ..utils import bold, green


def command_version():
    version = importlib.metadata.version("numerous.sdk")
    print(green(f"{bold('numerous.sdk')} command line tool v{bold(version)}"))



"""
This model takes input and returns an output multiplied by 2.
"""


def entrypoint(tag, system):
    class Model:
        def multiply_by_two(self, data, t):
            v = data.get("input1")
            output = {f"{tag}_output1": v * 2}
            return output

    return Model()



import warnings

from .numerous_admin_client import NumerousAdminClient, open_admin_client
from .numerous_client import (
    DataSourcesReader,
    DataStreamTimeoutBreakStatus,
    NumerousClient,
    ScenarioStatus,
    open_client,
)


def _setup_runtime():
    import signal
    import sys

    import urllib3

    from numerous.client.common import log

    urllib3.disable_warnings()

    def handle_term(_signum, _frame):
        log.warning("Terminated")
        sys.exit()

    signal.signal(signal.SIGINT, handle_term)
    signal.signal(signal.SIGTERM, handle_term)


_setup_runtime()


__all__ = (
    "DataSourcesReader",
    "NumerousClient",
    "open_client",
    "NumerousAdminClient",
    "open_admin_client",
    "ScenarioStatus",
    "DataStreamTimeoutBreakStatus",
)

warnings.warn(
    DeprecationWarning(
        "numerous.client has been deprecated in v1.0.0 and will be removed in a future release."
    )
)



import logging
from functools import wraps
from time import sleep

import grpc

log = logging.getLogger(__name__)


def _grpc_backoff(
    start_delay: float, attempt_number: int, max_attempts: int, rpc_error: grpc.RpcError
):
    if rpc_error.code() in (
        grpc.StatusCode.UNAVAILABLE,
        grpc.StatusCode.INTERNAL,
        grpc.StatusCode.UNKNOWN,
    ):
        if attempt_number == max_attempts:
            log.warning("Exceeded command retry attempts")
            raise rpc_error
        timeout = start_delay * 2**attempt_number
        log.warning(
            f"Got error: {rpc_error}, wait for {timeout} seconds and retry the command"
        )
        sleep(timeout)
    else:
        raise rpc_error


def grpc_retry_stream(max_attempts: int = 5, start_delay: float = 1.0):
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            for attempt_number in range(1, max_attempts + 1):
                try:
                    yield from func(self, *args, **kwargs)
                    return
                except grpc.RpcError as e:
                    _grpc_backoff(start_delay, attempt_number, max_attempts, e)

        return wrapper

    return decorator


def grpc_retry_unary(max_attempts: int = 5, start_delay: float = 1.0):
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            for attempt_number in range(1, max_attempts + 1):
                try:
                    return func(self, *args, **kwargs)
                except grpc.RpcError as e:
                    _grpc_backoff(start_delay, attempt_number, max_attempts, e)

        return wrapper

    return decorator


grpc_retry = grpc_retry_unary  # TODO: deprecate this for grpc_retry_unary



import os
from dataclasses import dataclass
from typing import Optional


@dataclass
class Config:
    max_status_message_length: int = 32
    min_status_update_interval_seconds: float = 1.0
    grpc_max_message_size: int = 4 * 1024 * 1024
    terminate_handler_timeout: Optional[float] = 5.0

    @staticmethod
    def from_environment():
        config = Config()
        if "MAX_STATUS_MESSAGE_LENGTH" in os.environ:
            config.max_status_message_length = int(
                os.environ["MAX_STATUS_MESSAGE_LENGTH"]
            )
        if "GRPC_MAX_MESSAGE_SIZE" in os.environ:
            config.grpc_max_message_size = int(os.environ["GRPC_MAX_MESSAGE_SIZE"])
        return config



import signal
import threading

from numerous.sdk.connect.hibernation import HibernationHandler
from numerous.sdk.connect.subscription import Subscription


class CommandHandler:
    def __init__(
        self,
        subscription: Subscription,
        hibernation_handler: HibernationHandler,
    ):
        self._subscription = subscription
        self._thread = threading.Thread(target=self._handler, daemon=True)
        self._thread.start()
        self._hibernation_handler = hibernation_handler

    def _handler(self):
        for message in self._subscription:
            command = message.message["command"]
            if command == "terminate":
                signal.raise_signal(signal.SIGTERM)
                return
            elif command == "hibernate":
                self._hibernation_handler.hibernate()
                return

    def close(self) -> None:
        self._thread.join()



import json
import signal
from datetime import datetime
from typing import Any, Callable, Optional

from numerous.grpc import spm_pb2, spm_pb2_grpc

from numerous.client_common.grpc_retry import grpc_retry_unary
from numerous.sdk.connect.job_state import JobState
from numerous.sdk.connect.job_utils import JobIdentifier


class HibernationHandler:
    def __init__(
        self,
        spm_client: spm_pb2_grpc.SPMStub,
        identity: JobIdentifier,
        execution_id: str,
        job_time_getter: Callable[[], Optional[datetime]],
        state_getter: Callable[[], JobState],
    ) -> None:
        self._spm_client = spm_client
        self._job_time_getter = job_time_getter
        self._identity = identity
        self._execution_id = execution_id
        self._state_getter = state_getter
        self.last_hibernate_time = self._load_last_hibernate_time()
        self.hibernating = False
        self.callback: Optional[Callable[[], Any]] = None

    def hibernate(self) -> None:
        self.hibernating = True
        if self.callback is not None:
            self.callback()
        self._state_getter().commit()
        self._store_last_hibernate_time()
        signal.raise_signal(signal.SIGTERM)

    @grpc_retry_unary()
    def _store_last_hibernate_time(self) -> None:
        job_time = self._job_time_getter()

        if job_time is None:
            return

        self._spm_client.SetScenarioCustomMetaData(
            request=spm_pb2.ScenarioCustomMetaData(
                project=self._identity.project_id,
                scenario=self._identity.scenario_id,
                execution=self._execution_id,
                key=f"{self._identity.job_id}.last_hibernate_job_time",
                meta=json.dumps(job_time.isoformat()),
            )
        )

    @grpc_retry_unary()
    def _load_last_hibernate_time(self) -> Optional[datetime]:
        last_hibernate_time = self._spm_client.GetScenarioCustomMetaData(
            request=spm_pb2.ScenarioCustomMetaData(
                project=self._identity.project_id,
                scenario=self._identity.scenario_id,
                execution=self._execution_id,
                key=f"{self._identity.job_id}.last_hibernate_job_time",
            )
        ).meta

        if not last_hibernate_time:
            return None

        try:
            return datetime.fromisoformat(json.loads(last_hibernate_time))
        except (ValueError, TypeError, json.decoder.JSONDecodeError):
            return None



"""Jobs represent the things that run on the numerous platform,
including the context in which code using the SDK is running.

This module contains the definition of :class:`Job` and the implementation
of its instances' creation.
"""

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional

from numerous.grpc.spm_pb2_grpc import FileManagerStub

from numerous.sdk.connect.status_handler import StatusHandler
from numerous.sdk.models.job_time import JobTime
from numerous.sdk.models.parameter import Parameter, get_parameters_dict


@dataclass
class Job:
    """A job in a :class:`numerous.sdk.models.scenario.Scenario`"""

    id: str
    name: str
    is_main: bool
    parameters: dict[str, Parameter]
    time: JobTime

    @staticmethod
    def from_document(
        job_id: str,
        data: dict[str, Any],
        file_manager_client: FileManagerStub,
        status_handler: StatusHandler,
        last_hibernate_time: Optional[datetime] = None,
    ) -> "Job":
        return Job(
            id=job_id,
            name=data.get("name", ""),
            is_main=data.get("isMain", False),
            parameters=get_parameters_dict(
                data.get("image", {}).get("parameters", []), file_manager_client
            ),
            time=JobTime.from_document(data, last_hibernate_time, status_handler),
        )



from enum import IntEnum
from getpass import getpass
from typing import Optional, Tuple

import grpc
from numerous.grpc.spm_pb2 import (
    AuthenticateUserRequest,
    RefreshUserRequest,
    TokenRequest,
)
from numerous.grpc.spm_pb2_grpc import TokenManagerStub

from .repository import NumerousRepository
from .utils import bold, get_token_manager, yellow


class AccessLevel(IntEnum):
    ANY = 0
    READ = 1
    WRITE = 2
    DEVELOPER = 3
    OWNER = 4
    ADMIN = 5


class AuthenticationError(Exception):
    pass


class IdTokenAuth(grpc.AuthMetadataPlugin):
    def __init__(self, id_token: str):
        self._id_token = id_token

    def __call__(
        self,
        _context: grpc.AuthMetadataContext,
        callback: grpc.AuthMetadataPluginCallback,
    ):
        callback([("authorization", self._id_token)], None)


def authenticate_user(
    token_manager: TokenManagerStub, email: str, password: str
) -> Tuple[str, str]:
    request = AuthenticateUserRequest(email=email, password=password)
    try:
        response = token_manager.AuthenticateUser(request)
        return response.id_token, response.refresh_token
    except grpc.RpcError:
        raise AuthenticationError()


def refresh_user(
    token_manager: TokenManagerStub, refresh_token: str
) -> Tuple[str, str, str]:
    request = RefreshUserRequest(refresh_token=refresh_token)
    try:
        response = token_manager.RefreshUser(request)
        return response.id_token, response.refresh_token, response.user_id
    except grpc.RpcError:
        raise AuthenticationError()


def get_refresh_token(repo: NumerousRepository, execution_id: str) -> str:
    if repo.remote is None:
        raise RuntimeError("Repository remote not defined")

    if repo.scenario is None:
        raise RuntimeError("Repository scenario is not defined")

    with get_token_manager(
        repo.remote.api_url, repo.remote.organization
    ) as token_manager:
        if repo.refresh_token is None:
            raise RuntimeError("Repository has no refresh token, use: numerous login")
        id_token, _, user_id = refresh_user(token_manager, repo.refresh_token)
        call_credentials = grpc.metadata_call_credentials(IdTokenAuth(id_token))
        request = TokenRequest(
            project_id=repo.scenario.project_id,
            scenario_id=repo.scenario.id,
            admin=False,
            execution_id=execution_id,
            job_id=repo.remote.job_id,
            access_level=AccessLevel.WRITE.value,
            user_id=user_id,
        )
        response, _ = token_manager.CreateRefreshToken.with_call(
            request, credentials=call_credentials
        )
        return str(response.val)


def get_id_token(api_url: str, refresh_token: str, organization: str) -> str:
    with get_token_manager(api_url, organization) as token_manager:
        return refresh_user(token_manager, refresh_token)[0]


def login(repo: NumerousRepository) -> str:
    if not repo.remote:
        raise AuthenticationError("No remote configured")

    id_token: Optional[str] = None

    with get_token_manager(
        repo.remote.api_url, repo.remote.organization
    ) as token_manager:
        if repo.refresh_token:
            try:
                id_token, refresh_token, _ = refresh_user(
                    token_manager, repo.refresh_token
                )
            except AuthenticationError:
                pass

        if id_token is None:
            print(yellow(f"Login to {bold(repo.remote)}"))
            email = input("Email: ")  # nosec
            password = getpass("Password: ")
            id_token, refresh_token = authenticate_user(token_manager, email, password)

    repo.refresh_token = refresh_token
    repo.save()

    return id_token



from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import Any, Generator, Optional, Union

from numerous.sdk.connect.status_handler import StatusHandler


class RunMode(Enum):
    CONTINUOUS = "continuous"
    DURATION = "duration"


@dataclass
class JobTime:
    """Contains information about the job runtime."""

    run_mode: RunMode
    start: datetime
    end: Optional[datetime]
    duration: Optional[timedelta]
    status_handler: StatusHandler
    current_time: Optional[datetime] = field(default=None, init=False)
    _last_hibernate_time: Optional[datetime] = field(default=None)

    @property
    def elapsed(self) -> Optional[timedelta]:
        if self.current_time is None:
            return None
        else:
            return self.current_time - self.start

    def start_to_time(self, t: datetime) -> timedelta:
        return t - self.start

    @staticmethod
    def from_document(
        job_data: dict[str, Any],
        last_hibernate_time: Optional[datetime],
        status_handler: StatusHandler,
    ) -> "JobTime":
        run_settings: dict[str, Any] = job_data["runSettings"]
        run_mode = RunMode(run_settings["runMode"])
        start = datetime.fromisoformat(run_settings["startDate"])
        end = (
            datetime.fromisoformat(run_settings["endDate"])
            if run_settings.get("endDate") is not None
            else None
        )
        return JobTime(
            run_mode=run_mode,
            status_handler=status_handler,
            start=start,
            end=end,
            duration=None if end is None else end - start,
            _last_hibernate_time=last_hibernate_time,
        )

    def range(
        self,
        step: Union[timedelta, float] = 1.0,
        start: Optional[datetime] = None,
        update_progress: Optional[bool] = False,
    ) -> Generator[datetime, None, None]:
        """Range over the runtime of the job, from :attr:`start` to :attr:`end`,
        inclusive. If :attr:`end` is not defined, range forever. If the job is resumed
        from hibernation, the start time will be the job time before hibernation.

        :param step: The step in time between yielded time value. If it is a `float` it
            is interpreted as seconds.
        :param start: If specified, it is used instead of the job start time.
        """
        if self.current_time is not None:
            raise RuntimeError("You can only range over job time once")

        step_ = step if isinstance(step, timedelta) else timedelta(seconds=step)
        if start is not None:
            self.current_time = start
        elif self._last_hibernate_time is not None:
            self.current_time = self._last_hibernate_time
        else:
            self.current_time = self.start
        while self.end is None or self.current_time <= self.end:
            if (
                update_progress
                and self.duration is not None
                and self.status_handler is not None
            ):
                self.status_handler.progress = (
                    100
                    * (self.current_time - self.start).total_seconds()
                    / self.duration.total_seconds()
                )
            yield self.current_time
            self.current_time += step_



import sys
from datetime import datetime
from logging import Logger
from pathlib import Path
from typing import List

from prettytable import PrettyTable

from ..auth import AuthenticationError, login
from ..build_manager_client import HistoryUpdate, get_build_manager_client
from ..repository import NumerousRepository
from ..utils import bold, cyan, green, print_load_warnings, red, yellow


def print_updates_table(repo: NumerousRepository, updates: List[HistoryUpdate]):
    table = PrettyTable()
    table.field_names = ["Timestamp", "Scenario", "Comment", "Commit ID", "Current"]
    for update in updates:
        table.add_row(
            [
                datetime.fromtimestamp(update.timestamp),
                update.scenario_id,
                update.comment,
                update.commit_id,
                f"{green('*')} {red('dirty') if repo.is_dirty() else ''}"
                if repo.commit == update.commit_id
                else "",
            ]
        )
    print(table)


def print_updates_git(repo: NumerousRepository, updates: List[HistoryUpdate]):
    for update in updates:
        state = (
            f" {green('current')}{red(' dirty') if repo.is_dirty() else ''}"
            if repo.commit == update.commit_id
            else ""
        )
        print(
            f"{yellow(f'commit {update.commit_id}')} {yellow('(')}"
            f"{cyan(update.project_id + ':' + update.scenario_id)}{yellow(')')}{state}"
        )
        update_time = datetime.fromtimestamp(update.timestamp)
        print(f"Author: {update.user_full_name} <{update.user_email}>")
        print(f'Date:   {update_time.strftime("%a %b %d %H:%M:%S %Y %z")}')
        print()
        for line in update.comment.splitlines():
            print("    " + line)
        print()


def command_log(log: Logger, path: Path, display_mode: str):
    path = path or Path.cwd()

    try:
        repo, load_warnings = NumerousRepository(path).load()
        print_load_warnings(load_warnings)
    except Exception as excp:
        log.debug(f"Exception {type(excp)} {excp}")
        print(red(f"Cannot get logs: {bold(str(path))} is not a repository."))
        sys.exit(1)

    if repo.remote is None:
        print(red("Cannot get logs: No remote is configured for the repository."))
        print(
            f"Use the command {bold('numerous config --remote <REMOTE_URL>')} to configure a remote for the "
            "repository."
        )
        sys.exit(1)

    if repo.scenario is None:
        print(red("Cannot get logs: A scenario must be checked out."))
        print(f"Check out a scenario with {bold('numerous checkout <SCENARIO ID>')}")
        print(
            f"Push to a scenario with {bold('numerous push -c <COMMENT> <SCENARIO ID>')}"
        )
        sys.exit(1)

    try:
        id_token = login(repo)
    except AuthenticationError:
        print(red("Cannot get logs: Login failed."))
        sys.exit(1)

    with get_build_manager_client(
        repo.remote.api_url, id_token, repo.remote.organization
    ) as build_manager:
        updates = build_manager.get_scenario_history(repo.remote.name, repo.scenario.id)

    if not updates:
        print(f"No history exists for remote {bold(repo.remote)}")

    if display_mode == "table":
        print_updates_table(repo, updates)
    else:
        print_updates_git(repo, updates)



from dataclasses import dataclass
from typing import Any

from numerous.grpc.spm_pb2_grpc import FileManagerStub

from .parameter import Parameter, get_parameters_dict


@dataclass
class Group:
    id: str
    name: str
    parameters: dict[str, Parameter]

    @staticmethod
    def from_document(
        data: dict[str, Any], file_manager_client: FileManagerStub, group_id
    ) -> "Group":
        return Group(
            id=group_id,
            name=data.get("groupName", ""),
            parameters=get_parameters_dict(
                data.get("parameters", []), file_manager_client=file_manager_client
            ),
        )



import json
import sys
from datetime import datetime
from logging import Logger, getLogger
from pathlib import Path
from typing import Optional

import requests

from ..auth import AuthenticationError, login
from ..build_manager_client import (
    SnapshotAlreadyExistsError,
    SnapshotError,
    get_build_manager_client,
)
from ..repository import NumerousRepository
from ..utils import (
    GzippedTarball,
    bold,
    cyan,
    green,
    hash_directory,
    print_load_warnings,
    red,
    yellow,
)

log = getLogger(__name__)
UNIX_VIRTUAL_ENV_DIRECTORY = "bin"
UNIX_VIRTUAL_ENV_FILES = {"python", "activate"}
WINDOWS_VIRTUAL_ENV_DIRECTORY = "Scripts"
WINDOWS_VIRTUAL_ENV_FILES = {"python.exe", "Activate.ps1"}


def get_snapshot_id(repo: NumerousRepository) -> str:
    return hash_directory(repo.path)


def get_snapshot_state(repo: NumerousRepository) -> dict[str, list[str]]:
    return {
        "files": [
            str(file_path.relative_to(repo.path))
            for file_path in repo.walk()
            if file_path.is_file()
        ]
    }


def print_push_error(msg: str, help_text: Optional[str] = None) -> None:
    print(red(f"Cannot push: {msg}"))
    if help_text:
        print(help_text)


def upload_snapshot(
    repo: NumerousRepository, upload_url: str, snapshot_id: str
) -> None:
    print("Uploading snapshot...")
    with GzippedTarball(repo.path, repo.walk()) as tar_path:
        log.debug(f"Created tarball {tar_path} for {repo.path}")
        with open(tar_path, "rb") as tar_file:
            upload_response = requests.post(
                upload_url,
                data=tar_file,
                headers={"Content-Type": "multipart/related"},
                params={"name": snapshot_id, "mimeType": "application/octet-stream"},
            )
            if upload_response.ok:
                log.debug(
                    f"Uploaded snapshot {snapshot_id} got response: {upload_response.status_code} "
                    f"{upload_response.reason}"
                )
            else:
                log.debug(
                    f"Failed uploading {snapshot_id} {upload_response.status_code} {upload_response.reason}"
                )
                raise RuntimeError("An error occured during upload")


def command_push(log: Logger, repo_path: Optional[Path], comment: str) -> None:
    path = Path.cwd() if repo_path is None else repo_path

    if not path.exists():
        print_push_error(f"{bold(path)} does not exist")
        sys.exit(1)

    repo, load_warnings = NumerousRepository(path).load()
    print_load_warnings(load_warnings)

    if _is_virtual_environment_included(repo):
        print_push_error(
            "Virtual environment hasn't been excluded, please use .exclude file to exclude it"
        )
        sys.exit(1)

    if repo.scenario is None:
        print_push_error("A scenario must be checked out.")
        sys.exit(1)

    if repo.snapshot == hash_directory(path):
        print_push_error("Cannot push: There is nothing new to push.")
        sys.exit(1)

    if repo.remote is None:
        help_text = (
            f'Use the command {bold("numerous config --remote <REMOTE_URL>")} to configure a remote for the '
            "repository."
        )
        print_push_error("No remote is configured for the repository.", help_text)
        sys.exit(1)

    try:
        id_token = login(repo)
    except AuthenticationError:
        print_push_error("Login failed.")
        sys.exit(1)

    print(cyan(f"Pushing to remote repository {bold(repo.remote)}..."))
    print("Creating snapshot metadata...")
    snapshot_id = get_snapshot_id(repo)
    snapshot_state = json.dumps(get_snapshot_state(repo))

    if len(snapshot_state) > 4e6:
        print_push_error(
            "The snapshot contains too many files and directories - please use the .exclude file to "
            "exclude files that are not needed."
        )
        sys.exit(1)

    with get_build_manager_client(
        repo.remote.api_url, id_token, repo.remote.organization
    ) as build_manager:
        try:
            upload_url = build_manager.create_snapshot_get_upload_url(
                repo.remote.name, snapshot_id, snapshot_state
            )
        except SnapshotAlreadyExistsError:
            print(yellow(f"Snapshot {snapshot_id} already exists. Stopping."))
            sys.exit(0)
        except SnapshotError:
            print_push_error("An error occured during snapshot creation.")
            sys.exit(1)

        try:
            upload_snapshot(repo, upload_url, snapshot_id)
        except Exception as e:
            print_push_error(str(e))
            sys.exit(1)

        commit_id = build_manager.create_history_update(
            repo.remote.name,
            repo.scenario.project_id,
            repo.scenario.id,
            snapshot_id,
            repo.commit,
            comment,
            datetime.utcnow().timestamp(),
        )
    print(green(f"Pushed snapshot {bold(snapshot_id)} as commit {bold(commit_id)}."))

    repo.commit = commit_id
    repo.snapshot = snapshot_id
    repo.save()


def _is_virtual_environment_included(repo: NumerousRepository) -> bool:
    for path in repo.walk():
        if not path.is_dir():
            continue

        if _is_unix_virtual_environment(path):
            return True
        elif _is_windows_virtual_environment(path):
            return True

    return False


def _is_unix_virtual_environment(path: Path):
    return path.name == UNIX_VIRTUAL_ENV_DIRECTORY and any(
        file_path.name in UNIX_VIRTUAL_ENV_FILES for file_path in path.iterdir()
    )


def _is_windows_virtual_environment(path: Path):
    return path.name == WINDOWS_VIRTUAL_ENV_DIRECTORY and any(
        file_path.name in WINDOWS_VIRTUAL_ENV_FILES for file_path in path.iterdir()
    )



from typing import Any, NoReturn


class DataFrameStub:
    def __call__(self, *args, **kwargs) -> NoReturn:
        raise LibraryNotInstalledError("pandas")

    def __getattr__(self, attr) -> NoReturn:  # noqa: F841
        raise LibraryNotInstalledError("pandas")


class PandasStub:
    def __getattr__(self, attr: str) -> Any:  # noqa: F841
        if attr == "DataFrame":
            return DataFrameStub()
        raise LibraryNotInstalledError("pandas")


class LibraryNotInstalledError(Exception):
    def __init__(self, library_name):
        self.library_name = library_name
        self.message = (
            f"{library_name} is not installed. "
            f"Please install it using 'pip install {library_name}'."
        )



import json
from contextlib import contextmanager
from datetime import datetime
from typing import Generator, Optional

from numerous.grpc import health_pb2, health_pb2_grpc, spm_pb2, spm_pb2_grpc

from . import config
from .common import RepeatedFunction, initialize_grpc_channel, parse_api_url


class NumerousAdminClient:
    def __init__(
        self,
        server: Optional[str] = None,
        port: Optional[int] = None,
        secure: Optional[bool] = None,
        refresh_token: Optional[str] = None,
        url: Optional[str] = None,
    ):
        self._refresh_token = refresh_token or config.NUMEROUS_API_REFRESH_TOKEN
        if self._refresh_token is None:
            raise RuntimeError("Refresh token must be set.")
        if url:
            server, port, secure = parse_api_url(url)
        self.channel, self._instance_id = initialize_grpc_channel(
            server=server or config.NUMEROUS_API_SERVER,
            port=port or config.NUMEROUS_API_PORT,
            refresh_token=self._refresh_token,
            secure=secure if secure is not None else not config.FORCE_INSECURE,
            token_callback=self._get_access_token,
        )
        self._base_client = spm_pb2_grpc.SPMStub(self.channel)
        self._job_manager = spm_pb2_grpc.JobManagerStub(self.channel)
        self._build_manager = spm_pb2_grpc.BuildManagerStub(self.channel)
        self._token_manager = spm_pb2_grpc.TokenManagerStub(self.channel)
        self._health = health_pb2_grpc.HealthStub(self.channel)

        self._access_token = None
        self._terminated = False
        self._complete = False
        self._user = "test_user"  # TODO: replace or remove the hardcoded value
        self._organization = (
            "EnergyMachines"  # TODO: replace or remove the hardcoded value
        )
        self._access_token_refresher = RepeatedFunction(
            interval=9 * 60,
            function=self._refresh_access_token,
            run_initially=True,
            refresh_token=self._refresh_token,
        )
        self._access_token_refresher.start()

    def _refresh_access_token(self, refresh_token):
        token = self._token_manager.GetAccessToken(
            spm_pb2.RefreshRequest(refresh_token=spm_pb2.Token(val=refresh_token))
        )
        self._access_token = token.val

    def _get_access_token(self):
        return self._access_token

    def launch_job(self, project, scenario, job) -> str:
        return str(
            self._job_manager.StartJob(
                spm_pb2.Job(
                    project_id=project,
                    scenario_id=scenario,
                    job_id=job,
                    user_id=self._user,
                )
            ).execution_id
        )

    def terminate_job(self, project, scenario, job):
        self._job_manager.TerminateJob(
            spm_pb2.Job(
                project_id=project,
                scenario_id=scenario,
                job_id=job,
                user_id=self._user,
            )
        )

    def reset_job(self, project, scenario, job):
        self._job_manager.ResetJob(
            spm_pb2.Job(
                project_id=project,
                scenario_id=scenario,
                job_id=job,
                user_id=self._user,
            )
        )

    def update_execution(self, update_):
        self._base_client.UpdateExecution(spm_pb2.Json(json=json.dumps(update_)))

    def get_execution_status(self, exe_id):
        return self._job_manager.GetExecutionStatus(
            spm_pb2.ExecutionId(execution_id=exe_id)
        ).json

    def clear_execution(self, scenario, execution):
        self._base_client.ClearExecutionMemory(
            spm_pb2.Scenario(scenario=scenario, execution=execution)
        )

    def remove_data_execution(self, scenario, execution):
        self._base_client.RemoveExecutionData(
            spm_pb2.Scenario(scenario=scenario, execution=execution)
        )

    def remove_execution(self, scenario, execution):
        self.remove_data_execution(scenario, execution)
        self.remove_execution_schedule(execution)

    def delete_execution(self, exe_id):
        self._job_manager.DeleteExecution(spm_pb2.ExecutionId(execution_id=exe_id))

    def update_job_by_backend(
        self, project, scenario, job, exe, message, status, log=None, complete=False
    ):
        self._base_client.SetScenarioProgress(
            spm_pb2.ScenarioProgress(
                project=project,
                scenario=scenario,
                job_id=job,
                message=message,
                status=status,
                progress=0,
            )
        )

        if log is not None:
            self._base_client.PushExecutionLogEntries(
                spm_pb2.LogEntries(
                    scenario=scenario,
                    execution_id=exe,
                    log_entries=[log],
                    timestamps=[datetime.utcnow().timestamp()],
                )
            )

        if complete:
            self._base_client.CompleteExecutionIgnoreInstance(
                spm_pb2.ExecutionMessage(
                    project_id=project,
                    scenario_id=scenario,
                    job_id=job,
                    execution_id=exe,
                )
            )

        return spm_pb2.Empty()

    def read_execution_logs(self, execution_id, project_id):
        for log_entry in self._base_client.ReadExecutionLogEntries(
            spm_pb2.ExecutionReadLogs(
                execution_id=execution_id, project_id=project_id, start=0, end=0
            )
        ):
            yield log_entry.log_entry, datetime.fromtimestamp(log_entry.timestamp)

    def hibernate_job(self, project_id, scenario_id, job_id, organization_id, user_id):
        self._job_manager.HibernateJob(
            spm_pb2.Job(
                project_id=project_id,
                scenario_id=scenario_id,
                job_id=job_id,
                user_id=user_id,
            )
        )

    def resume_job(self, project_id, scenario_id, job_id, organization_id, user_id):
        self._job_manager.ResumeJob(
            spm_pb2.Job(
                project_id=project_id,
                scenario_id=scenario_id,
                job_id=job_id,
                user_id=user_id,
            )
        )

    def add_job_schedule(
        self,
        project_id,
        scenario_id,
        job_id,
        organization_id,
        user_id,
        sleep_after,
        sleep_for,
        enable_scheduling,
    ):
        # TODO: check schedule attributes
        self._job_manager.AddJobSchedule(
            spm_pb2.JobSchedule(
                job=spm_pb2.Job(
                    project_id=project_id,
                    scenario_id=scenario_id,
                    job_id=job_id,
                    organization_id=organization_id,
                    user_id=user_id,
                ),
                schedule=spm_pb2.Schedule(
                    sleep_after=sleep_after,
                    sleep_for=sleep_for,
                    enable_scheduling=enable_scheduling,
                ),
            )
        )

    def update_job_schedule(
        self,
        project_id,
        scenario_id,
        job_id,
        organization_id,
        user_id,
        sleep_after,
        sleep_for,
        enable_scheduling,
    ):
        self._job_manager.AddJobSchedule(
            spm_pb2.JobSchedule(
                job=spm_pb2.Job(
                    project_id=project_id,
                    scenario_id=scenario_id,
                    job_id=job_id,
                    organization_id=organization_id,
                    user_id=user_id,
                ),
                schedule=spm_pb2.Schedule(
                    sleep_after=sleep_after,
                    sleep_for=sleep_for,
                    enable_scheduling=enable_scheduling,
                ),
            )
        )

    def remove_job_schedule(
        self, project_id, scenario_id, job_id, organization_id, user_id
    ):
        self._job_manager.RemoveJobSchedule(
            spm_pb2.Job(
                project_id=project_id,
                scenario_id=scenario_id,
                job_id=job_id,
                organization_id=organization_id,
                user_id=user_id,
            )
        )

    def get_job_schedule(
        self, project_id, scenario_id, job_id, organization_id, user_id
    ):
        return self._job_manager.GetJobSchedule(
            spm_pb2.Job(
                project_id=project_id,
                scenario_id=scenario_id,
                job_id=job_id,
                organization_id=organization_id,
                user_id=user_id,
            )
        )

    def add_execution_schedule(
        self, execution_id, sleep_after, sleep_for, enable_scheduling
    ):
        self._job_manager.AddExecutionSchedule(
            spm_pb2.ExecutionAndSchedule(
                execution=spm_pb2.ExecutionId(execution_id=execution_id),
                schedule=spm_pb2.Schedule(
                    sleep_after=sleep_after,
                    sleep_for=sleep_for,
                    enable_scheduling=enable_scheduling,
                ),
            )
        )

    def update_execution_schedule(
        self, execution_id, sleep_after, sleep_for, enable_scheduling
    ):
        self._job_manager.UpdateExecutionSchedule(
            spm_pb2.ExecutionAndSchedule(
                execution=spm_pb2.ExecutionId(execution_id=execution_id),
                schedule=spm_pb2.Schedule(
                    sleep_after=sleep_after,
                    sleep_for=sleep_for,
                    enable_scheduling=enable_scheduling,
                ),
            )
        )

    def remove_execution_schedule(self, execution_id):
        self._job_manager.RemoveExecutionSchedule(
            spm_pb2.ExecutionId(execution_id=execution_id)
        )

    def get_execution_schedule(self, execution_id):
        return self._job_manager.GetExecutionSchedule(
            spm_pb2.ExecutionId(execution_id=execution_id)
        )

    def generate_scenario_token(
        self,
        project,
        scenario,
        job,
        user_id,
        organization_id,
        agent,
        purpose,
        admin=False,
    ):
        token = self._token_manager.CreateRefreshToken(
            spm_pb2.TokenRequest(
                project_id=project,
                scenario_id=scenario,
                job_id=job,
                user_id=user_id,
                organization_id=organization_id,
                agent=agent,
                purpose=purpose,
                admin=admin,
            )
        ).val

        return token

    def enqueue_build(self, repository, commit_id, launch_after_build=False, tags=None):
        resp = self._build_manager.LaunchBuild(
            spm_pb2.BuildRequest(
                repository=repository,
                commit_id=commit_id,
                launch_after_build=launch_after_build,
                tags=tags,
            )
        )
        return resp.build_id

    def update_scenario_image(
        self,
        name,
        path,
        project,
        scenario,
        job,
        repository,
        commit,
        build,
        internal=True,
    ):
        self._build_manager.UpdateJobImageReference(
            spm_pb2.JobImageReferenceUpdate(
                project=project,
                scenario=scenario,
                job=job,
                name=name,
                path=path,
                repository=repository,
                commit_id=commit,
                internal=internal,
                build=build,
            )
        )

    def update_build(self, repository, build_id, **update):
        self._build_manager.UpdateBuild(
            spm_pb2.BuildUpdate(
                repository=repository, build_id=build_id, update=json.dumps(update)
            )
        )

    def get_build_info(self, repository, build_id):
        response = self._build_manager.GetBuildInfo(
            spm_pb2.BuildInfoRequest(repository=repository, build_id=build_id)
        )
        return json.loads(response.info)

    def push_scenario_logs_admin(self, logs, exe_id):
        timestamps = [log_entries[0].timestamp() for log_entries in logs]

        self._base_client.PushExecutionLogEntries(
            spm_pb2.LogEntries(
                scenario="",
                execution_id=exe_id,
                log_entries=[log_entries[1] for log_entries in logs],
                timestamps=timestamps,
            )
        )

    def delete_system(self, system):
        self._base_client.DeleteSystem(spm_pb2.SystemRequest(system=system))

    def delete_user(self, user):
        self._base_client.DeleteUser(spm_pb2.UserRequest(user=user))

    def complete_signup(self, invitationID, firstName, lastName, organization):
        self._base_client.CompleteUserSignup(
            spm_pb2.SignupData(
                invitationID=invitationID,
                firstName=firstName,
                lastName=lastName,
                organization=organization,
            )
        )

    def login_get_refresh(self, email, password):
        return self._token_manager.LoginGetRefreshToken(
            spm_pb2.LoginCredentials(
                email=email,
                password=password,
                token_request=spm_pb2.TokenRequest(
                    project_id="ssfsdf",
                    scenario_id="sdffd",
                    admin=True,
                    execution_id="sdsfd",
                    job_id="sfdsdf",
                    user_id="sfsdfsdfsdf",
                    organization_id="sdfsfd",
                    access_level=4,
                ),
            )
        )

    def check_health(self):
        response = self._health.Check(health_pb2.HealthCheckRequest(service=""))
        return response.status == health_pb2.HealthCheckResponse.ServingStatus.SERVING

    def close(self):
        self.channel.close()
        self._access_token_refresher.stop()


@contextmanager
def open_admin_client(
    server: Optional[str] = None,
    port: Optional[int] = None,
    secure: Optional[bool] = None,
    refresh_token: Optional[str] = None,
) -> Generator[NumerousAdminClient, None, None]:
    numerous_client_ = NumerousAdminClient(
        server=server, port=port, secure=secure, refresh_token=refresh_token
    )

    try:
        yield numerous_client_

    finally:
        numerous_client_.close()



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Any, Optional, Union


class InputSourceType(Enum):
    SCENARIO = "scenario"
    DATASET = "dataset"


@dataclass
class InputSource:
    project_id: str
    scenario_id: str
    type: InputSourceType
    offset: float = 0.0

    def job_time_to_base_time(self, time: datetime) -> float:
        return time.timestamp() - self.offset


@dataclass
class InputVariableScenarioSource:
    source: InputSource
    tag: str
    value: Optional[float] = field(default=None, init=False)
    """Should only be set by `InputReader`"""


@dataclass
class InputVariableStaticSource:
    value: float


def get_input_sources_from_scenario_document(
    data: dict[str, Any]
) -> dict[str, InputSource]:
    input_sources: list[dict[str, Any]] = data.get("inputScenarios", [])
    return {
        input_source["scenarioID"]: InputSource(
            project_id=input_source["projectID"],
            scenario_id=input_source["scenarioID"],
            type=InputSourceType(input_source["type"]),
            offset=input_source.get("offset", 0.0),
        )
        for input_source in input_sources
    }


@dataclass
class InputVariable:
    uuid: str
    id: str
    path: list[str]
    display_name: str
    source: Union[InputVariableScenarioSource, InputVariableStaticSource]
    scale: float = 1.0
    offset: float = 0.0

    @property
    def key(self) -> str:
        return ".".join(self.path)

    @property
    def value(self) -> Optional[float]:
        return self.source.value



import traceback


def tb_str(e):
    s = traceback.format_exception(e, value=e, tb=e.__traceback__)
    es = [ss.strip() for ss in s]
    return "\n".join(es)


class SimulationError(Exception):
    pass



import json
import logging
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Iterator, List, Optional

import grpc
from numerous.grpc import spm_pb2
from numerous.grpc.spm_pb2_grpc import BuildManagerStub, SPMStub

from .auth import IdTokenAuth
from .utils import OrganizationInterceptor, StubClient

log = logging.getLogger(__name__)


class SnapshotAlreadyExistsError(Exception):
    pass


class BuildError(Exception):
    pass


class SnapshotError(Exception):
    pass


class HistoryUpdate:
    def __init__(self, update: spm_pb2.HistoryUpdate):
        self.commit_id: str = update.id
        self.snapshot_id: str = update.snapshot_id
        self.scenario_id: str = update.scenario_id
        self.project_id: str = update.project_id
        self.comment: str = update.comment
        self.timestamp: float = update.timestamp
        self.user_full_name: str = update.user.full_name
        self.user_email: str = update.user.email


class BuildManagerClient:
    def __init__(self, stub: BuildManagerStub, spm_stub: SPMStub):
        self._stub: BuildManagerStub = stub
        self._spm_stub = spm_stub

    def launch_build(
        self,
        repository: str,
        commit_id: str,
        launch_after_build: bool,
        tags: List[str],
        timeout: Optional[float],
    ):
        request = spm_pb2.BuildRequest(
            repository=repository,
            commit_id=commit_id,
            launch_after_build=launch_after_build,
            tags=tags,
            timeout=timeout or 0.0,
        )
        response = self._stub.LaunchBuild(request)
        return response.build_id

    def get_build_info(self, repository: str, build_id: str) -> Any:
        request = spm_pb2.BuildInfoRequest(repository=repository, build_id=build_id)
        response = self._stub.GetBuildInfo(request)
        return json.loads(response.info)

    def update_job_image_reference(
        self,
        repository: str,
        project_id: str,
        scenario_id: str,
        job_id: str,
        path: str,
        commit_id: str,
        build_id: str,
    ):
        request = spm_pb2.JobImageReferenceUpdate(
            project=project_id,
            scenario=scenario_id,
            job=job_id,
            name=repository,
            path=path,
            repository=repository,
            commit_id=commit_id,
            internal=True,
            build=build_id,
        )
        self._stub.UpdateJobImageReference(request)

    def get_scenario_history(
        self, repository: str, scenario_id: str
    ) -> List[HistoryUpdate]:
        request = spm_pb2.HistoryRequest(repository=repository, scenario_id=scenario_id)
        response = self._stub.GetHistory(request)
        return [HistoryUpdate(update) for update in response.updates]

    def get_history_update(
        self, repository: str, commit_id: str
    ) -> Optional[HistoryUpdate]:
        request = spm_pb2.HistoryRequest(
            repository=repository, history_update_ids=[commit_id]
        )
        response = self._stub.GetHistory(request)
        return HistoryUpdate(response.updates[0]) if len(response.updates) > 0 else None

    def create_history_update(
        self,
        repository: str,
        project_id: str,
        scenario_id: str,
        snapshot_id: str,
        parent_commit_id: Optional[str],
        comment: str,
        timestamp: float,
    ) -> str:
        request = spm_pb2.HistoryUpdate(
            repository=repository,
            project_id=project_id,
            scenario_id=scenario_id,
            snapshot_id=snapshot_id,
            parent_commit_id=parent_commit_id or "",
            comment=comment,
            timestamp=datetime.utcnow().timestamp(),
        )
        response = self._stub.CreateHistoryUpdate(request)
        return str(response.commit_id)

    def get_download_url(
        self, repository: str, project_id: str, scenario_id: str, snapshot_id: str
    ) -> str:
        request = spm_pb2.PullRequest(
            repository=repository,
            snapshot_id=snapshot_id,
            project_id=project_id,
            scenario_id=scenario_id,
        )
        response = self._stub.PullSnapshot(request)
        return str(response.archive_url.url)

    def create_snapshot_get_upload_url(
        self, repository: str, snapshot_id: str, file_structure: Any
    ) -> str:
        request = spm_pb2.Snapshot(
            snapshot_id=snapshot_id,
            repository=repository,
            file_structure=file_structure,
            only_get_uploads=False,
        )

        try:
            response = self._stub.CreateSnapshot(request)
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.ALREADY_EXISTS:
                raise SnapshotAlreadyExistsError()
            else:
                raise SnapshotError()

        return str(response.upload_url)

    def stream_build_logs(
        self, build_id: str, scenario_id: str, project_id: str
    ) -> Iterator[str]:
        request = spm_pb2.Subscription(
            channel_patterns=[build_id], project_id=project_id, scenario=scenario_id
        )
        for update in self._spm_stub.SubscribeForUpdates(request):
            if update.message == ":BUILD_DONE:":
                break
            elif update.message == ":ERROR:":
                raise BuildError()
            else:
                yield update.message


@contextmanager
def get_build_manager_client(
    api_url: str, id_token: str, organization: str
) -> Iterator[BuildManagerClient]:
    with StubClient(
        api_url,
        SPMStub,
        IdTokenAuth(id_token),
        [OrganizationInterceptor(organization)],
    ) as spm_stub:
        with StubClient(
            api_url,
            BuildManagerStub,
            IdTokenAuth(id_token),
            [OrganizationInterceptor(organization)],
        ) as stub:
            yield BuildManagerClient(stub, spm_stub)



"""Functionality related to reading data and the :class:`Reader`."""

from collections import defaultdict
from typing import Generator, Optional, Union

from numerous.sdk.utils import PandasStub

try:
    import pandas
except ModuleNotFoundError:
    pandas = PandasStub()

from numerous.grpc import spm_pb2, spm_pb2_grpc

from numerous.sdk.connect.job_utils import JobIdentifier


class Reader:
    """The :class:`Reader` reads data from the server."""

    def __init__(
        self,
        spm_stub: spm_pb2_grpc.SPMStub,
        job_identity: JobIdentifier,
        execution_id: str,
    ):
        """Initialize the reader.

        :param spm_stub: The client.
        :param job_identity: The identity of the job, the :class:`Reader` is reading for.
        :param execution_id: The ID of the execution, the :class:`Reader` is reading for.
        """
        self._spm_stub = spm_stub
        self._job_identity = job_identity
        self._execution_id = execution_id

    def range(
        self, start: Optional[float], end: Optional[float]
    ) -> Generator[dict[str, list[float]], None, None]:
        """Reads a range of data and returns an iterable of data rows."""
        yield from self._data_stream(start or 0.0, end or 0.0)

    def data_frame(
        self, start: Optional[float], end: Optional[float]
    ) -> pandas.DataFrame:
        """Reads a range of data and returns a `pandas.DataFrame` with the data."""
        data = defaultdict(list)
        for data_list in self._data_stream(start or 0.0, end or 0.0):
            for key, values in data_list.items():
                data[key].extend(values)

        return pandas.DataFrame({k: v for k, v in data.items()})

    def _data_stream(
        self,
        start: float = 0.0,
        end: float = 0.0,
        tags: Union[list[str], None] = None,
        subscribe: bool = False,
    ) -> Generator[dict[str, list[float]], None, None]:
        time_range = False
        if start is not None and end is not None and start >= 0 and end > 0:
            time_range = True

        raw_data_stream = self._spm_stub.ReadData(
            spm_pb2.ReadScenario(
                project=self._job_identity.project_id,
                scenario=self._job_identity.scenario_id,
                execution=self._execution_id,
                tags=tags,
                start=start or 0.0,
                end=end or 0.0,
                time_range=time_range,
                listen=subscribe,
            )
        )

        for data_list in raw_data_stream:
            data: dict[str, list[float]] = defaultdict(list)
            for data_block in data_list.data:
                data[data_block.tag].extend(data_block.values)
            yield data



from .build import BuildArgs, command_build  # noqa: F401
from .checkout import command_checkout  # noqa: F401
from .clone import command_clone  # noqa: F401
from .config import command_config  # noqa: F401
from .init import command_init  # noqa: F401
from .log import command_log  # noqa: F401
from .login import command_login  # noqa: F401
from .push import command_push  # noqa: F401
from .version import command_version  # noqa: F401



import enum
import threading
from datetime import datetime
from queue import Empty, Queue
from time import time
from typing import Any, Dict, List, Optional

import grpc

from .common import Interp, interp1d, log, trace


class DataSourceEmptyDataset(Exception):
    pass


class DataSourceHibernating(Exception):
    pass


class DataSourceCompleted(Exception):
    pass


class DataSourceStreamClosed(Exception):
    pass


class DataSourceStreamStarting(Exception):
    pass


class SourceStatus(enum.Enum):
    EMPTY_DATASET = 0
    HIBERNATING = 1
    COMPLETED = 2
    STREAM_CLOSED = 3
    OK = 4
    NO_DATA = 5
    STREAM_STARTING = 6


class StaticSource:
    def __init__(
        self, row_data, t0=0, t_end=None, dt=3600, generate_index=True  # noqa: F841
    ):
        self.row = row_data
        self.source_type = "static"
        self.t = float("inf")
        self.t0 = t0
        self.subscribe = False
        self.generate_index = generate_index

    def get_at_time(self, t):
        self.t = t
        self.row["_index"] = t
        self.row["_index_relative"] = t - self.t0
        return self.row


class DynamicSource:
    def __init__(
        self,
        client: Any,
        scenario,
        execution,
        spec,
        t0=0,
        te=0,
        timeout=None,
        source_type=None,
        interpolation_method=Interp.zero,
    ):
        self.timeout = timeout
        self.t0 = t0
        if te is None:
            te = 0
        self.te = te
        self.meta = client.get_timeseries_meta_data(
            scenario=scenario, execution=execution
        )
        self.closed = client.get_dataset_closed(scenario=scenario, execution=execution)
        self.name = scenario
        self.spec = spec
        self.spec["offset"] = self.spec["offset"] if "offset" in self.spec else 0

        # The offset set in spec is the full offset from 0 - not relative from meta.offset
        offset = self.spec["offset"]
        self.offset = offset
        self.execution = execution
        self.client = client
        self.logger = client.logger  # for trace decorator
        self.trace = client.trace  # for trace decorator

        # start time in original absolute time/input base time
        start = t0 - offset
        end = max(te - offset, 0)

        # If the data set is open, we will try to listen for data as its being generated if t end not reached.
        # This is stored in the state, to enable consistent behaviour across hibernation resumes
        subscribe = client.state.get(f"_{scenario}_{execution}_source_subscribe", None)
        if subscribe is None:
            subscribe = not self.closed
            client.state.set(f"_{scenario}_{execution}_source_subscribe", subscribe)
        self.subscribe = subscribe

        stats = client.get_timeseries_stats(None, scenario, execution)
        if stats.min + offset > t0:
            raise Exception(
                f"dynamic input scenario {scenario} with tags {[t.name for t in self.meta.tags]} "
                f"is offset to the future {stats.min + offset}>{t0}"
            )

        self.start = start
        self.end = end

        log.info("Scenario: %s", scenario)
        log.info("Scenario data is closed: %s", self.closed)
        log.info("Offset in spec: %s", datetime.fromtimestamp(spec["offset"]))
        log.info("Offset in meta: %s", datetime.fromtimestamp(self.meta.offset))
        log.info(
            "Data stats: %s",
            client.get_timeseries_stats(
                project=None, scenario=scenario, execution=execution
            ),
        )
        log.info("Combined offset: %s", datetime.fromtimestamp(offset))
        log.info("t0: %s", datetime.fromtimestamp(t0))
        log.info("te: %s", datetime.fromtimestamp(te))
        log.info("start: %s (%s)", datetime.fromtimestamp(start), start)
        log.info("end: %s (%s)", datetime.fromtimestamp(end), end)

        tags = set([s["tag"] for s in spec["tags"].values()])
        self.tags = tags

        self.generator = client.read_data_as_row(
            scenario=scenario,
            tags=tags,
            execution=execution,
            start=start,
            end=end,
            subscribe=self.subscribe,
            index_by_time=True,
        )
        self.source_type = source_type
        self.row_queue: Queue = Queue(100)
        self.t = -1
        self._rows_read = 0  # Keep a counter of rows read - if nothing is read we don't want to requery.
        self.row_buffer = RowBuffer(
            tags=[s for s in spec["tags"].keys()],
            offset=offset,
            method=interpolation_method,
        )
        # Keep the last/presumably max ix read on record - if in continuous mode this needs to be  added to the requery
        # as an offset.
        self.max_ix_read = 0
        self.last_get: Optional[float] = None
        self.status = SourceStatus.STREAM_STARTING

        def _read_producer_thread():
            try:
                _read_producer()
            except grpc.RpcError:
                self.row_queue.put("CLOSED")
            except Exception as error:
                self.row_queue.put(error)

        def _read_producer():
            complete = False
            lastdt = 0
            lastrow_index = None
            status = SourceStatus.STREAM_CLOSED

            while not complete:
                for row_ in self.generator:
                    if row_ == "HIBERNATE":
                        status = SourceStatus.HIBERNATING
                        break
                    elif row_ == "STOP":
                        status = SourceStatus.COMPLETED
                        break

                    self._rows_read += 1
                    self.row_queue.put(row_)
                    if lastrow_index is None:
                        lastrow_index = row_["_index"]
                    lastdt = row_["_index"] - lastrow_index
                    lastrow_index = row_["_index"]
                    self.max_ix_read = row_["_index"]

                log.debug(
                    "%s: All data read: %s of %d to %d, read n rows: %d",
                    scenario,
                    self.t - offset,
                    start,
                    end,
                    self._rows_read,
                )

                if self._rows_read <= 0:
                    # If nothing was returned the data source is empty and no use to requery.
                    log.debug(f"No data in {scenario}!")
                    self.row_queue.put("EMPTY")
                    return
                elif self.subscribe:
                    if status == SourceStatus.STREAM_CLOSED:
                        log.debug(
                            f"Subscribed data source {scenario} is active and not flagged hibernating, "
                            f"but is not streaming"
                        )
                        self.row_queue.put("CLOSED")
                    elif status == SourceStatus.COMPLETED:
                        log.debug(
                            f"Subscribed data source {scenario} has closed and stopped generating data"
                        )
                        self.row_queue.put("STOP")
                    elif status == SourceStatus.HIBERNATING:
                        log.debug(f"Subscribed data source {scenario} is hibernating")
                        self.row_queue.put("HIBERNATE")
                    return
                else:
                    # mode is continuous requry data and shift its _index to extend the current data
                    log.debug("Requerying data for " + scenario)
                    # TODO: Assess if we need to keep the query in memory instead of requery.
                    # Requery could lead to inconsistency perhaps? Data changed since last query?
                    self.generator = client.read_data_as_row(
                        tags=tags,
                        scenario=scenario,
                        execution=execution,
                        start=self.start,
                        end=self.end,
                        # We should never subscribe on requery
                        subscribe=False,
                        # Make sure to offset requeried _index so we extend the data. Shift by one estimated timestep
                        offset=self.max_ix_read + lastdt - self.start,
                    )

                if self.end > 0:
                    complete = self.max_ix_read >= self.end

            self.row_queue.put("STOP")

        threading.Thread(target=_read_producer_thread, daemon=True).start()

    def __next__(self):
        if self.status not in (SourceStatus.OK, SourceStatus.STREAM_STARTING):
            return self.status

        if self.last_get is None:
            self.last_get = time()
        try:
            row_ = self.row_queue.get(block=True, timeout=1)
            self.last_get = time()
        except Empty:
            row_ = None
            if self.timeout is not None and time() - self.last_get > self.timeout:
                log.debug("timeout")
                raise TimeoutError("Timed out!")

        if row_ == "STOP":
            log.debug("source consumed and no longer active")
            self.status = SourceStatus.COMPLETED
            return self.status
        elif row_ == "EMPTY":
            self.status = SourceStatus.EMPTY_DATASET
            log.debug("Source empty")
            return self.status
        elif row_ == "HIBERNATE":
            self.status = SourceStatus.HIBERNATING
            log.debug("Source is hibernating")
            return self.status
        elif row_ == "CLOSED":
            self.status = SourceStatus.STREAM_CLOSED
            log.debug("Source unexpectedly closed")
            return self.status
        elif row_ is None:
            if self.status != SourceStatus.STREAM_STARTING:
                self.status = SourceStatus.NO_DATA
            return self.status
        elif isinstance(row_, Exception):
            raise row_

        row = {}
        for tag, s in self.spec["tags"].items():
            try:
                row[tag] = row_[s["tag"]] * s["scale"] + s["offset"]
            except Exception:
                log.debug(str(row_))
                raise
        # substract meta.offset since read_as_row returns the data in input base time
        row["_index"] = row_["_index"] + self.offset
        row["_index_relative"] = row_["_index"] - self.t0
        row["_datetime_utc"] = datetime.utcfromtimestamp(row["_index"])
        self.t = row["_index"]
        self.row_buffer.add(row)
        return SourceStatus.OK

    # Get at simulation time
    @trace
    def get_at_time(self, t):
        while self.t <= t or not self.row_buffer.full:
            status = self.__next__()
            if status != SourceStatus.OK:
                break

        if t > self.row_buffer.tmax():
            if self.status == SourceStatus.EMPTY_DATASET:
                raise DataSourceEmptyDataset(f"Data Source for {self.name} is empty.")
            elif self.status == SourceStatus.HIBERNATING:
                raise DataSourceHibernating(
                    f"Data Source for {self.name} is hibernating."
                )
            elif self.status == SourceStatus.COMPLETED:
                raise DataSourceCompleted(
                    f"Data Source for {self.name} has been consumed and is no longer active."
                )
            elif self.status == SourceStatus.STREAM_CLOSED:
                raise DataSourceStreamClosed(
                    f"Data Source for {self.name} closed it's stream unexpectedly."
                )
            elif self.status == SourceStatus.NO_DATA:
                return None
            elif self.status == SourceStatus.STREAM_STARTING:
                raise DataSourceStreamStarting(
                    f"Data Source for {self.name} is starting"
                )
            else:
                raise NotImplementedError(f"unhandled exception for {self.status}")
        else:
            return self.row_buffer.get(t)


class FIFO:
    def __init__(self, buffer_len: int = 2):
        self.buffer: Dict[str, List[float]] = {}
        self.buffer_len = buffer_len

    def _init_buffer(self, row):
        self.buffer = {tag: [0.0] * self.buffer_len for tag in row.keys()}

    def __len__(self):
        return len(self.buffer)

    def add(self, row):
        if len(self.buffer) == 0:
            self._init_buffer(row)

        for tag, val in row.items():
            self.buffer[tag].pop(0)
            self.buffer[tag].append(val)


class RowBuffer:
    def __init__(self, tags=None, offset=None, method=Interp.zero):
        self.tags = tags
        self.tags.append("_index")
        self.offset = offset
        self.method = method
        self.full = False
        self._rows = 0
        if method == Interp.zero:
            self.buffer_len = 2
        elif method == Interp.linear:
            self.buffer_len = 2
        else:
            raise NotImplementedError
        self._buffer = FIFO(self.buffer_len)

    def add(self, row):
        row_tags = {tag: row.get(tag) for tag in self.tags}
        self._buffer.add(row_tags)

        if self.full:
            return

        self._rows += 1
        if self._rows >= self.buffer_len:
            self.full = True

    def tmax(self):
        if "_index" in self._buffer.buffer:
            return max(self._buffer.buffer["_index"])
        else:
            return -1

    def get(self, t):
        tags = self.tags
        irow = {}
        if self.method == Interp.zero:
            for tag in tags[:-1]:
                ifun = interp1d(
                    self._buffer.buffer["_index"], self._buffer.buffer[tag], kind="zero"
                )
                irow.update({tag: ifun(t)[0]})
        elif self.method == Interp.linear:
            for tag in tags[:-1]:
                ifun = interp1d(
                    self._buffer.buffer["_index"],
                    self._buffer.buffer[tag],
                    kind="linear",
                )
                irow.update({tag: ifun(t)[0]})
        else:
            raise NotImplementedError("Unknown interpolation type")
        return irow


class InputManager:
    def __init__(self, sources, t, t0):
        self.sources = sources
        self.t = t
        self.t0 = t0

    def get_at_time(self, t):
        row = {}
        for source in self.sources:
            try:
                part_row = source.get_at_time(t)
                if part_row is None:
                    return None

            except TimeoutError as e:
                if source.subscribe:
                    return None
                else:
                    raise e
            row.update(part_row)
        # return simulation time
        row.update({"_index": t})
        return row



from dataclasses import dataclass
from typing import Any

from numerous.grpc.spm_pb2_grpc import FileManagerStub

from .parameter import Parameter, get_parameters_dict


@dataclass
class Project:
    id: str
    name: str
    parameters: dict[str, Parameter]

    @staticmethod
    def from_document(
        data: dict[str, Any], file_manager_client: FileManagerStub
    ) -> "Project":
        return Project(
            id=data["id"],
            name=data["projectName"],
            parameters=get_parameters_dict(
                data.get("parameters", []), file_manager_client
            ),
        )
